# Copyright (C) 2025 Alexander Vanhee
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Any
from gi.repository import Gtk, Gdk, Graphene, Gsk, GObject
import math
import cairo

class CropOverlay(Gtk.Widget):
    __gtype_name__ = "GradiaCropOverlay"

    interactive = GObject.Property(
        type=bool,
        default=False,
        nick="Interactive",
        blurb="Whether the crop overlay allows user interaction"
    )

    aspect_ratio = GObject.Property(
        type=float,
        default=0.0,
        nick=_("_Aspect Ratio"),
        blurb="Locked aspect ratio (width/height). 0 means unlocked, 1 means square"
    )

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.picture_widget: Gtk.Picture | None = None

        self.crop_x = 0.0
        self.crop_y = 0.0
        self.crop_width = 1.0
        self.crop_height = 1.0

        self.handle_size = 30
        self.edge_grab_distance = 8
        self.dragging_handle = None
        self.dragging_edge = None
        self.dragging_area = False
        self.drag_start_x = 0
        self.drag_start_y = 0
        self.drag_start_crop = None

        self.gesture_drag = Gtk.GestureDrag()
        self.gesture_drag.connect("drag-begin", self._on_drag_begin)
        self.gesture_drag.connect("drag-update", self._on_drag_update)
        self.gesture_drag.connect("drag-end", self._on_drag_end)
        self.add_controller(self.gesture_drag)

        self.motion_controller = Gtk.EventControllerMotion()
        self.motion_controller.connect("motion", self._on_motion)
        self.add_controller(self.motion_controller)

        self.set_cursor_from_name("crosshair")

        self.connect("notify::interactive", self._on_interactive_changed)
        self.connect("notify::aspect-ratio", self._on_aspect_ratio_changed)

    def _on_interactive_changed(self, obj, pspec):
        if not self.interactive:
            self.dragging_handle = None
            self.dragging_edge = None
            self.dragging_area = False
            self.set_cursor_from_name("default")
        else:
            self.set_cursor_from_name("crosshair")

        self.queue_draw()

    def _on_aspect_ratio_changed(self, obj, pspec):
        self._apply_aspect_ratio()
        self.queue_draw()

    def _apply_aspect_ratio(self) -> None:
        if self.aspect_ratio <= 0.0 or not self.picture_widget:
            return

        img_x, img_y, img_w, img_h = self._get_image_bounds()
        if img_w <= 0 or img_h <= 0:
            return

        target_rel_aspect = (self.aspect_ratio * img_h) / img_w
        current_rel_aspect = self.crop_width / self.crop_height

        if abs(target_rel_aspect - current_rel_aspect) < 1e-5:
            return

        center_x = self.crop_x + self.crop_width / 2
        center_y = self.crop_y + self.crop_height / 2

        if self.crop_width / target_rel_aspect > self.crop_height:
            new_width = self.crop_height * target_rel_aspect
            self.crop_x = center_x - new_width / 2
            self.crop_width = new_width
        else:
            new_height = self.crop_width / target_rel_aspect
            self.crop_y = center_y - new_height / 2
            self.crop_height = new_height

        self._clamp_crop_rectangle()

    def do_snapshot(self, snapshot: Gtk.Snapshot) -> None:
        if not self.picture_widget or not self.picture_widget.get_paintable() or not self.has_crop():
            return

        width = self.get_width()
        height = self.get_height()
        if width <= 0 or height <= 0:
            return

        img_x, img_y, img_w, img_h = self._get_image_bounds()
        crop_x = img_x + (self.crop_x * img_w)
        crop_y = img_y + (self.crop_y * img_h)
        crop_w = self.crop_width * img_w
        crop_h = self.crop_height * img_h

        self._draw_background_overlay(snapshot, img_x, img_y, img_w, img_h, crop_x, crop_y, crop_w, crop_h)
        if self.interactive:
            self._draw_inner_border(snapshot, Gdk.RGBA(red=1.0, green=1.0, blue=1.0, alpha=0.8) , crop_x, crop_y, crop_w, crop_h)
            self._draw_corner_lines(snapshot, crop_x, crop_y, crop_w, crop_h)
        else:
            self._draw_inner_border(snapshot, Gdk.RGBA(red=1.0, green=1.0, blue=1.0, alpha=0.5) , crop_x, crop_y, crop_w, crop_h)

    def _draw_background_overlay(self, snapshot: Gtk.Snapshot, img_x: float, img_y: float, img_w: float, img_h: float,
                                 crop_x: float, crop_y: float, crop_w: float, crop_h: float) -> None:
        overlay_color = Gdk.RGBA(red=0.2, green=0.2, blue=0.2, alpha=0.6)

        if crop_y > img_y:
            top_overlay = Graphene.Rect.alloc()
            top_overlay.init(img_x-1, img_y-1, img_w+1, crop_y - img_y +1)
            snapshot.append_color(overlay_color, top_overlay)

        if crop_y + crop_h < img_y + img_h:
            bottom_overlay = Graphene.Rect.alloc()
            bottom_overlay.init(img_x-1, crop_y + crop_h, img_w +1, (img_y + img_h) - (crop_y + crop_h))
            snapshot.append_color(overlay_color, bottom_overlay)

        if crop_x > img_x:
            left_overlay = Graphene.Rect.alloc()
            left_overlay.init(img_x-1, crop_y - 0.20, crop_x - img_x +1, crop_h + 0.4)
            snapshot.append_color(overlay_color, left_overlay)

        if crop_x + crop_w < img_x + img_w:
            right_overlay = Graphene.Rect.alloc()
            right_overlay.init(crop_x + crop_w, crop_y - 0.20, (img_x + img_w) - (crop_x + crop_w), crop_h + 0.4)
            snapshot.append_color(overlay_color, right_overlay)

    def _draw_inner_border(self, snapshot: Gtk.Snapshot, border_color, crop_x: float, crop_y: float, crop_w: float, crop_h: float) -> None:
        border_width = 2.0

        top_border = Graphene.Rect.alloc()
        top_border.init(crop_x, crop_y - border_width/2, crop_w, border_width)
        snapshot.append_color(border_color, top_border)

        bottom_border = Graphene.Rect.alloc()
        bottom_border.init(crop_x, crop_y + crop_h - border_width/2, crop_w, border_width)
        snapshot.append_color(border_color, bottom_border)

        left_border = Graphene.Rect.alloc()
        left_border.init(crop_x - border_width/2, crop_y, border_width, crop_h)
        snapshot.append_color(border_color, left_border)

        right_border = Graphene.Rect.alloc()
        right_border.init(crop_x + crop_w - border_width/2, crop_y, border_width, crop_h)
        snapshot.append_color(border_color, right_border)

    def _draw_corner_lines(self, snapshot: Gtk.Snapshot, crop_x: float, crop_y: float, crop_w: float, crop_h: float) -> None:
        corner_color = Gdk.RGBA(red=1.0, green=1.0, blue=1.0, alpha=1.0)
        corner_line_width = 4.0
        corner_line_length = 25.0
        border_width = 0.0
        offset = border_width

        top_left_h = Graphene.Rect.alloc()
        top_left_h.init(crop_x - corner_line_width - offset, crop_y - offset - corner_line_width, corner_line_length + corner_line_width + offset, corner_line_width)
        snapshot.append_color(corner_color, top_left_h)

        top_left_v = Graphene.Rect.alloc()
        top_left_v.init(crop_x - offset - corner_line_width, crop_y - corner_line_width - offset, corner_line_width, corner_line_length + corner_line_width + offset)
        snapshot.append_color(corner_color, top_left_v)

        top_right_h = Graphene.Rect.alloc()
        top_right_h.init(crop_x + crop_w - corner_line_length, crop_y - offset - corner_line_width, corner_line_length + corner_line_width + offset, corner_line_width)
        snapshot.append_color(corner_color, top_right_h)

        top_right_v = Graphene.Rect.alloc()
        top_right_v.init(crop_x + crop_w + offset, crop_y - corner_line_width - offset, corner_line_width, corner_line_length + corner_line_width + offset)
        snapshot.append_color(corner_color, top_right_v)

        bottom_left_h = Graphene.Rect.alloc()
        bottom_left_h.init(crop_x - corner_line_width - offset, crop_y + crop_h + offset, corner_line_length + corner_line_width + offset, corner_line_width)
        snapshot.append_color(corner_color, bottom_left_h)

        bottom_left_v = Graphene.Rect.alloc()
        bottom_left_v.init(crop_x - offset - corner_line_width, crop_y + crop_h - corner_line_length, corner_line_width, corner_line_length + corner_line_width + offset)
        snapshot.append_color(corner_color, bottom_left_v)

        bottom_right_h = Graphene.Rect.alloc()
        bottom_right_h.init(crop_x + crop_w - corner_line_length, crop_y + crop_h + offset, corner_line_length + corner_line_width + offset, corner_line_width)
        snapshot.append_color(corner_color, bottom_right_h)

        bottom_right_v = Graphene.Rect.alloc()
        bottom_right_v.init(crop_x + crop_w + offset, crop_y + crop_h - corner_line_length, corner_line_width, corner_line_length + corner_line_width + offset)
        snapshot.append_color(corner_color, bottom_right_v)

    def _get_handle_at_point(self, x: float, y: float) -> str | None:
        if not self.interactive or not self.picture_widget or not self.picture_widget.get_paintable():
            return None

        img_x, img_y, img_w, img_h = self._get_image_bounds()

        crop_x = img_x + (self.crop_x * img_w)
        crop_y = img_y + (self.crop_y * img_h)
        crop_w = self.crop_width * img_w
        crop_h = self.crop_height * img_h

        half_handle = self.handle_size / 2

        corners = {
            "top-left": (crop_x, crop_y),
            "top-right": (crop_x + crop_w, crop_y),
            "bottom-right": (crop_x + crop_w, crop_y + crop_h),
            "bottom-left": (crop_x, crop_y + crop_h),
        }

        for handle_name, (corner_x, corner_y) in corners.items():
            if (abs(x - corner_x) <= half_handle and
                abs(y - corner_y) <= half_handle):
                return handle_name

        return None

    def _get_edge_at_point(self, x: float, y: float) -> str | None:
        if not self.interactive or not self.picture_widget or not self.picture_widget.get_paintable():
            return None

        img_x, img_y, img_w, img_h = self._get_image_bounds()

        crop_x = img_x + (self.crop_x * img_w)
        crop_y = img_y + (self.crop_y * img_h)
        crop_w = self.crop_width * img_w
        crop_h = self.crop_height * img_h

        if self.aspect_ratio <= 0:
            edges = {
                "top": (crop_x + crop_w/2, crop_y),
                "right": (crop_x + crop_w, crop_y + crop_h/2),
                "bottom": (crop_x + crop_w/2, crop_y + crop_h),
                "left": (crop_x, crop_y + crop_h/2),
            }

            half_handle = self.handle_size / 2
            for edge_name, (edge_x, edge_y) in edges.items():
                if (abs(x - edge_x) <= half_handle and abs(y - edge_y) <= half_handle):
                    return edge_name

        grab_distance = self.edge_grab_distance

        if (crop_x <= x <= crop_x + crop_w and abs(y - crop_y) <= grab_distance):
            return "top"
        if (crop_x <= x <= crop_x + crop_w and abs(y - (crop_y + crop_h)) <= grab_distance):
            return "bottom"
        if (crop_y <= y <= crop_y + crop_h and abs(x - crop_x) <= grab_distance):
            return "left"
        if (crop_y <= y <= crop_y + crop_h and abs(x - (crop_x + crop_w)) <= grab_distance):
            return "right"

        return None

    def _is_point_in_crop_area(self, x: float, y: float) -> bool:
        if not self.interactive or not self.picture_widget or not self.picture_widget.get_paintable():
            return False

        img_x, img_y, img_w, img_h = self._get_image_bounds()

        crop_x = img_x + (self.crop_x * img_w)
        crop_y = img_y + (self.crop_y * img_h)
        crop_w = self.crop_width * img_w
        crop_h = self.crop_height * img_h

        if (crop_x <= x <= crop_x + crop_w and crop_y <= y <= crop_y + crop_h):
            return (self._get_handle_at_point(x, y) is None and
                    self._get_edge_at_point(x, y) is None)

        return False

    def _on_drag_begin(self, gesture: Gtk.GestureDrag, start_x: float, start_y: float) -> None:
        if not self.interactive:
            return

        handle = self._get_handle_at_point(start_x, start_y)
        if handle:
            self.dragging_handle = handle
            self.drag_start_crop = (self.crop_x, self.crop_y, self.crop_width, self.crop_height)
            gesture.set_state(Gtk.EventSequenceState.CLAIMED)
        else:
            edge = self._get_edge_at_point(start_x, start_y)
            if edge:
                self.dragging_edge = edge
                self.drag_start_crop = (self.crop_x, self.crop_y, self.crop_width, self.crop_height)
                gesture.set_state(Gtk.EventSequenceState.CLAIMED)
            elif self._is_point_in_crop_area(start_x, start_y):
                self.dragging_area = True
                self.drag_start_crop = (self.crop_x, self.crop_y, self.crop_width, self.crop_height)
                gesture.set_state(Gtk.EventSequenceState.CLAIMED)

        if self.dragging_handle or self.dragging_edge or self.dragging_area:
            self.drag_start_x = start_x
            self.drag_start_y = start_y

    def _on_drag_update(self, gesture: Gtk.GestureDrag, offset_x: float, offset_y: float) -> None:
        if not self.interactive:
            return

        if (self.dragging_handle or self.dragging_edge or self.dragging_area) and self.drag_start_crop:
            current_x = self.drag_start_x + offset_x
            current_y = self.drag_start_y + offset_y

            if self.dragging_handle:
                self._update_crop_from_handle_drag(current_x, current_y)
            elif self.dragging_edge:
                self._update_crop_from_edge_drag(current_x, current_y)
            elif self.dragging_area:
                self._update_crop_from_area_drag(current_x, current_y)
            self.queue_draw()

    def _on_drag_end(self, gesture: Gtk.GestureDrag, offset_x: float, offset_y: float) -> None:
        if not self.interactive:
            return

        self.dragging_handle = None
        self.dragging_edge = None
        self.dragging_area = False
        self.set_cursor_from_name("crosshair")

    def _on_motion(self, controller: Gtk.EventControllerMotion, x: float, y: float) -> None:
        if not self.interactive:
            return

        if not (self.dragging_handle or self.dragging_edge or self.dragging_area):
            self._update_cursor(x, y)

    def _update_cursor(self, x: float, y: float):
        handle = self._get_handle_at_point(x, y)
        if handle:
            if handle == "top-left":
                self.set_cursor_from_name("nw-resize")
            if handle == "top-right":
                self.set_cursor_from_name("ne-resize")
            if handle == "bottom-right":
                self.set_cursor_from_name("se-resize")
            if handle == "bottom-left":
                self.set_cursor_from_name("sw-resize")
            return

        edge = self._get_edge_at_point(x, y)
        if edge:
            if edge in ["top", "bottom"]:
                self.set_cursor_from_name("ns-resize")
            else:
                self.set_cursor_from_name("ew-resize")
        elif self._is_point_in_crop_area(x, y):
            self.set_cursor_from_name("move")
        else:
            self.set_cursor_from_name("crosshair")

    def _update_crop_from_handle_drag(self, x: float, y: float) -> None:
        if not self.drag_start_crop: return
        img_x, img_y, img_w, img_h = self._get_image_bounds()
        if img_w <= 0 or img_h <= 0: return

        dx = (x - self.drag_start_x) / img_w
        dy = (y - self.drag_start_y) / img_h
        start_x, start_y, start_w, start_h = self.drag_start_crop

        if self.aspect_ratio > 0.0:
            target_rel_aspect = (self.aspect_ratio * img_h) / img_w

            if self.dragging_handle == "top-left":
                anchor_x, anchor_y = start_x + start_w, start_y + start_h
                new_w = start_w - dx
                new_h = new_w / target_rel_aspect
                self.crop_x, self.crop_y = anchor_x - new_w, anchor_y - new_h
                self.crop_width, self.crop_height = new_w, new_h
            elif self.dragging_handle == "top-right":
                anchor_x, anchor_y = start_x, start_y + start_h
                new_w = start_w + dx
                new_h = new_w / target_rel_aspect
                self.crop_x, self.crop_y = anchor_x, anchor_y - new_h
                self.crop_width, self.crop_height = new_w, new_h
            elif self.dragging_handle == "bottom-left":
                anchor_x, anchor_y = start_x + start_w, start_y
                new_w = start_w - dx
                new_h = new_w / target_rel_aspect
                self.crop_x, self.crop_y = anchor_x - new_w, anchor_y
                self.crop_width, self.crop_height = new_w, new_h
            elif self.dragging_handle == "bottom-right":
                new_w = start_w + dx
                new_h = new_w / target_rel_aspect
                self.crop_width, self.crop_height = new_w, new_h
        else:
            if self.dragging_handle == "top-left": self.crop_x, self.crop_y, self.crop_width, self.crop_height = start_x + dx, start_y + dy, start_w - dx, start_h - dy
            elif self.dragging_handle == "top-right": self.crop_y, self.crop_width, self.crop_height = start_y + dy, start_w + dx, start_h - dy
            elif self.dragging_handle == "bottom-right": self.crop_width, self.crop_height = start_w + dx, start_h + dy
            elif self.dragging_handle == "bottom-left": self.crop_x, self.crop_width, self.crop_height = start_x + dx, start_w - dx, start_h + dy

        self._clamp_crop_rectangle()

    def _update_crop_from_edge_drag(self, x: float, y: float) -> None:
        if not self.drag_start_crop: return
        img_x, img_y, img_w, img_h = self._get_image_bounds()
        if img_w <= 0 or img_h <= 0: return

        dx = (x - self.drag_start_x) / img_w
        dy = (y - self.drag_start_y) / img_h
        start_x, start_y, start_w, start_h = self.drag_start_crop

        if self.aspect_ratio > 0.0:
            target_rel_aspect = (self.aspect_ratio * img_h) / img_w
            center_x, center_y = start_x + start_w / 2, start_y + start_h / 2

            if self.dragging_edge == "top":
                new_h = start_h - dy
                new_w = new_h * target_rel_aspect
                self.crop_x, self.crop_y = center_x - new_w / 2, start_y + dy
                self.crop_width, self.crop_height = new_w, new_h
            elif self.dragging_edge == "bottom":
                new_h = start_h + dy
                new_w = new_h * target_rel_aspect
                self.crop_x = center_x - new_w / 2
                self.crop_width, self.crop_height = new_w, new_h
            elif self.dragging_edge == "left":
                new_w = start_w - dx
                new_h = new_w / target_rel_aspect
                self.crop_x, self.crop_y = start_x + dx, center_y - new_h / 2
                self.crop_width, self.crop_height = new_w, new_h
            elif self.dragging_edge == "right":
                new_w = start_w + dx
                new_h = new_w / target_rel_aspect
                self.crop_y = center_y - new_h / 2
                self.crop_width, self.crop_height = new_w, new_h
        else:
            if self.dragging_edge == "top": self.crop_y, self.crop_height = start_y + dy, start_h - dy
            elif self.dragging_edge == "bottom": self.crop_height = start_h + dy
            elif self.dragging_edge == "left": self.crop_x, self.crop_width = start_x + dx, start_w - dx
            elif self.dragging_edge == "right": self.crop_width = start_w + dx

        self._clamp_crop_rectangle()

    def _update_crop_from_area_drag(self, x: float, y: float) -> None:
        if not self.drag_start_crop: return
        img_x, img_y, img_w, img_h = self._get_image_bounds()
        if img_w <= 0 or img_h <= 0: return

        dx = (x - self.drag_start_x) / img_w
        dy = (y - self.drag_start_y) / img_h
        start_x, start_y, _, _ = self.drag_start_crop

        self.crop_x = start_x + dx
        self.crop_y = start_y + dy

        self._clamp_crop_rectangle()

    def _clamp_crop_rectangle(self):
        min_dim = 0.2
        self.crop_width = max(min_dim, self.crop_width)
        self.crop_height = max(min_dim, self.crop_height)

        if self.crop_width > 1.0: self.crop_width = 1.0
        if self.crop_height > 1.0: self.crop_height = 1.0

        self.crop_x = max(0, min(1 - self.crop_width, self.crop_x))
        self.crop_y = max(0, min(1 - self.crop_height, self.crop_y))

    def set_picture_reference(self, picture: Gtk.Picture) -> None:
        self.picture_widget = picture
        if picture:
            picture.connect("notify::paintable", lambda *args: self.queue_draw())
            picture.connect("notify::paintable", lambda *args: self._apply_aspect_ratio())

    def _get_image_bounds(self) -> tuple[float, float, float, float]:
        if not self.picture_widget or not self.picture_widget.get_paintable():
            return 0, 0, self.get_width(), self.get_height()

        widget_width = self.picture_widget.get_width()
        widget_height = self.picture_widget.get_height()
        paintable = self.picture_widget.get_paintable()
        image_width = paintable.get_intrinsic_width()
        image_height = paintable.get_intrinsic_height()

        if image_width <= 0 or image_height <= 0:
            return 0, 0, widget_width, widget_height

        scale = min(widget_width / image_width, widget_height / image_height)
        display_width = image_width * scale
        display_height = image_height * scale

        offset_x = (widget_width - display_width) / 2
        offset_y = (widget_height - display_height) / 2

        return offset_x, offset_y, display_width, display_height

    def get_crop_rectangle(self) -> tuple[float, float, float, float]:
        return self.crop_x, self.crop_y, self.crop_width, self.crop_height

    def has_crop(self) -> bool:
        return self.crop_width != 1 or self.crop_height != 1

    def set_crop_rectangle(self, x: float, y: float, width: float, height: float) -> None:
        self.crop_x = x
        self.crop_y = y
        self.crop_width = width
        self.crop_height = height
        self._clamp_crop_rectangle()
        self.queue_draw()



# Copyright (C) 2025 Alexander Vanhee
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

def parse_aspect_ratio(text: str) -> float | None:
    text = text.strip()
    if not text:
        return None
    if ":" in text:
        num, denom = map(float, text.split(":"))
        if denom == 0:
            raise ValueError("Denominator cannot be zero")
        return num / denom
    return float(text)

def check_aspect_ratio_bounds(ratio: float, min_ratio=0.2, max_ratio=5) -> bool:
    return min_ratio <= ratio <= max_ratio



# Copyright (C) 2022 Gradience Team
# Copyright 2023-2025, tfuxu <https://github.com/tfuxu>
# Copyright (C) 2025 Alexander Vanhee
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
import sys
import traceback

from gradia.constants import build_type  # pyright: ignore

class Logger(logging.getLoggerClass()):
    """
    This is a wrapper of `logging` module. It provides
    custom formatting for log messages.

    Attributes
    ----------
    issue_footer_levels : list, optional
        Custom list of levels on which to show issue footer.
        [Allowed values: warning, error, traceback_error, critical]
    formatter : str, optional
        Custom formatter for the logger.
    """

    log_colors = {
        "debug": 32,
        "info": 36,
        "warning": 33,
        "error": 31,
        "critical": 41
    }

    log_format = {
        'fmt': '%(message)s'
    }

    issue_footer = "If you are reporting an issue, please copy the logs printed above to the issue body."

    issue_footer_levels = [
        "error",
        "traceback_error",
        "critical"
    ]

    def __set_traceback_info(self, exception: BaseException | None = None) -> str:
        if not exception:
            exception = sys.exc_info()[1]
        traceback = self.get_traceback(exception)

        message_head = "\n\t\033[1mTraceback:\033[0m"
        message_body = f"\n\033[90m{traceback}\033[0m"
        message_body = message_body.replace("\n", "\n\t\t")

        return message_head + message_body

    def __set_exception_info(self, exception: BaseException | None = None) -> str:
        if not exception:
            exception = sys.exc_info()[1]

        message_head = "\n\t\033[1mException:\033[0m"
        message_body = f"\n{exception}"
        message_body = message_body.replace("\n", "\n\t\t")

        return message_head + message_body

    def __set_level_color(self, level: str, message: str) -> str:
        color_id = self.log_colors[level]

        return f"[\033[1;{color_id}m{level.upper()}\033[0m] {message}"

    def __init__(self, issue_footer_levels: list | None = None, fmt: str | None = None) -> None:
        """
        The constructor for Logger class.
        """

        super().__init__(name="Gradia")

        if issue_footer_levels:
            self.issue_footer_levels = issue_footer_levels

        formatter = logging.Formatter(self.log_format["fmt"])
        if fmt:
            formatter = logging.Formatter(fmt)

        if build_type == "debug":
            self.root.setLevel(logging.DEBUG)
        else:
            self.root.setLevel(logging.INFO)

        self.root.handlers = []

        handler = logging.StreamHandler()
        handler.setFormatter(formatter)
        self.root.addHandler(handler)

    def debug(self, message: str, *args, **kwargs) -> None:
        self.root.debug(self.__set_level_color("debug", str(message)))

    def info(self, message: str, *args, **kwargs) -> None:
        self.root.info(self.__set_level_color("info", str(message)))

    def warning(self, message: str, exception: BaseException | None = None,
                    show_exception: bool = False, show_traceback: bool = False, *args, **kwargs) -> None:
        if show_exception:
            message += self.__set_exception_info(exception)
        if show_traceback:
            message += self.__set_traceback_info(exception)

        self.root.warning(self.__set_level_color("warning", str(message)))
        if "warning" in self.issue_footer_levels:
            self.print_issue_footer()

    def error(self, message: str, exception: BaseException | None = None,
                show_exception: bool = False, show_traceback: bool = False, *args, **kwargs) -> None:
        if show_exception:
            message += self.__set_exception_info(exception)
        if show_traceback:
            message += self.__set_traceback_info(exception)

        self.root.error(self.__set_level_color("error", str(message)))
        if "error" in self.issue_footer_levels:
            self.print_issue_footer()

    def traceback_error(self, message: str, exception: BaseException | None = None,
                            show_exception: bool = False) -> None:
        if show_exception:
            message += self.__set_exception_info(exception)
        message += self.__set_traceback_info(exception)

        self.root.error(self.__set_level_color("error", str(message)))
        if "traceback_error" in self.issue_footer_levels:
            self.print_issue_footer()

    def critical(self, message: str, exception: BaseException | None = None,
                    show_exception: bool = False, show_traceback: bool = True, *args, **kwargs) -> None:
        if show_exception:
            message += self.__set_exception_info(exception)
        if show_traceback:
            message += self.__set_traceback_info(exception)

        self.root.critical(self.__set_level_color("critical", str(message)))
        if "critical" in self.issue_footer_levels:
            self.print_issue_footer()

    def set_silent(self) -> None:
        self.root.handlers = []

    def print_issue_footer(self) -> None:
        self.root.info(self.__set_level_color("info", self.issue_footer))

    def get_traceback(self, exception: BaseException | None) -> str | None:
        if not exception:
            return

        traceback_list = traceback.format_exception(exception)
        exception_tb = "".join(traceback_list)

        return exception_tb


"""
Use for testing only.
How to execute: python -m gradia.backend.logger
"""
# pylint: disable=W0719,W0718,W0707
if __name__ == "__main__":
    logging = Logger()

    logging.info("This is an information.")
    logging.debug("This is a debug message.")

    try:
        raise ArithmeticError("Arithmetic Error")
    except Exception:
        try:
            raise Exception("General Exception")
        except Exception as e:
            logging.traceback_error("This is an test error.", exception=e, show_exception=True)

            print(f"Retrieved traceback: {logging.get_traceback(e)}")



# Copyright (C) 2025 Alexander Vanhee, tfuxu
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Adw, Gtk, Gio, GdkPixbuf, GLib
from typing import Callable, Optional, Any
import threading, os
from gradia.utils.colors import HexColor
from gradia.graphics.gradient import GradientBackground, Gradient
from gradia.app_constants import PREDEFINED_GRADIENTS, PRESET_IMAGES


class BasePresetButton(Gtk.MenuButton):
    def __init__(
        self,
        callback: Optional[Callable[[Any], None]] = None,
        icon_name: str = "view-grid-symbolic",
        **kwargs
    ) -> None:
        super().__init__(**kwargs)
        self.callback = callback
        self.set_icon_name(icon_name)
        self.set_valign(Gtk.Align.CENTER)
        self.add_css_class("flat")
        self._setup_popover()

    def _setup_popover(self) -> None:
        self.popover = Gtk.Popover()
        self.set_popover(self.popover)

        main_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        main_box.set_spacing(8)
        main_box.set_margin_top(8)
        main_box.set_margin_bottom(8)
        main_box.set_margin_start(8)
        main_box.set_margin_end(8)

        self.flowbox = Gtk.FlowBox()
        self.flowbox.set_max_children_per_line(3)
        self.flowbox.set_selection_mode(Gtk.SelectionMode.NONE)
        self.flowbox.set_homogeneous(True)
        self.flowbox.set_row_spacing(8)
        self.flowbox.set_column_spacing(8)
        self.flowbox.connect("child-activated", self._on_item_selected)

        self._create_preset_buttons()

        main_box.append(self.flowbox)
        self.popover.set_child(main_box)

    def _create_preset_buttons(self) -> None:
        raise NotImplementedError("Subclasses must implement _create_preset_buttons")

    def _on_item_selected(self, flowbox, item):
        raise NotImplementedError("Subclasses must implement _on_item_selected")

    def _create_bin(self, width: int = 60, height: int = 40, css_class: str = "preset") -> Adw.Bin:
        bin_widget = Adw.Bin()
        bin_widget.set_size_request(width, height)
        bin_widget.add_css_class(css_class)
        bin_widget.add_css_class("flat")
        return bin_widget

    def set_callback(self, callback: Callable[[Any], None]) -> None:
        self.callback = callback

    def _clear_flowbox(self) -> None:
        child = self.flowbox.get_first_child()
        while child:
            next_child = child.get_next_sibling()
            self.flowbox.remove(child)
            child = next_child


class GradientPresetButton(BasePresetButton):
    __gtype_name__ = "GradiaGradientPresetButton"

    def _create_preset_buttons(self) -> None:
        for i, gradient in enumerate(PREDEFINED_GRADIENTS[:6]):
            gradient_name = f"gradient-preset-{i}"
            bin_widget = self._create_bin(css_class="preset-button")
            bin_widget.set_name(gradient_name)
            self._apply_gradient_to_bin(bin_widget, gradient)

            self.flowbox.append(bin_widget)
            flowbox_child = bin_widget.get_parent()

            flowbox_child.update_property(
                [Gtk.AccessibleProperty.LABEL],
                [gradient.name]
            )

    def _apply_gradient_to_bin(self, bin_widget: Adw.Bin, gradient: Gradient) -> None:
        css = f"""
            .preset-button {{
                background-image: {gradient.to_css()};
            }}
        """
        css_provider = Gtk.CssProvider()
        css_provider.load_from_string(css)
        bin_widget.get_style_context().add_provider(
            css_provider,
            Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION + 3
        )

    def _on_item_selected(self, flowbox, item):
        index = item.get_index()
        if index < len(PREDEFINED_GRADIENTS[:6]):
            gradient = PREDEFINED_GRADIENTS[index]
            self.popover.popdown()
            if self.callback:
                self.callback(gradient)

    def set_gradient_presets(self, gradients: list[Gradient]) -> None:
        self._clear_flowbox()

        for i, gradient in enumerate(gradients[:6]):
            gradient_name = f"gradient-preset-{i}"
            bin_widget = self._create_bin(css_class="gradient-preset")
            bin_widget.set_name(gradient_name)
            self._apply_gradient_to_bin(bin_widget, gradient)
            self.flowbox.append(bin_widget)


class ImagePresetButton(BasePresetButton):
    __gtype_name__ = "GradiaImagePresetButton"

    def __init__(
        self,
        callback: Optional[Callable[[str], None]] = None,
        **kwargs
    ) -> None:
        super().__init__(callback=callback, **kwargs)

    def _get_image_label(self, path: str) -> str:
        filename = os.path.basename(path)
        name_without_ext = os.path.splitext(filename)[0]
        return name_without_ext.replace('-', ' ').replace('_', ' ').title()

    def _create_preset_buttons(self) -> None:
        for path in PRESET_IMAGES:
            bin_widget = self._create_bin(width=80, height=60, css_class="preset-button")

            picture = Gtk.Picture()
            picture.set_content_fit(Gtk.ContentFit.COVER)
            picture.set_halign(Gtk.Align.FILL)
            picture.set_valign(Gtk.Align.FILL)
            picture.set_hexpand(True)
            picture.set_vexpand(True)

            bin_widget.set_child(picture)
            self._load_preset_image_async(path, picture)

            self.flowbox.append(bin_widget)
            flowbox_child = bin_widget.get_parent()

            label = self._get_image_label(path)
            flowbox_child.update_property(
                [Gtk.AccessibleProperty.LABEL],
                [label]
            )

    def _on_item_selected(self, flowbox, item):
        index = item.get_index()
        if index < len(PRESET_IMAGES):
            path = PRESET_IMAGES[index]
            self.popover.popdown()
            if self.callback:
                self.callback(path)

    def _load_preset_image_async(self, resource_path: str, picture: Gtk.Picture) -> None:
        def load_in_background():
            try:
                resource = Gio.resources_lookup_data(resource_path, Gio.ResourceLookupFlags.NONE)
                data = resource.get_data()

                if data is None:
                    raise RuntimeError("Failed to get data from resource lookup")

                loader = GdkPixbuf.PixbufLoader.new()
                loader.write(data)
                loader.close()

                pixbuf = loader.get_pixbuf()

                if pixbuf is None:
                    raise RuntimeError("Failed to load pixbuf from resource data")

                GLib.idle_add(self._on_preset_image_loaded, picture, pixbuf)

            except Exception as e:
                print(f"Error loading preset image {resource_path}: {e}")
                GLib.idle_add(self._on_preset_image_error, picture)

        thread = threading.Thread(target=load_in_background, daemon=True)
        thread.start()

    def _on_preset_image_loaded(self, picture: Gtk.Picture, pixbuf: GdkPixbuf.Pixbuf) -> bool:
        try:
            max_width, max_height = 80, 60

            width = pixbuf.get_width()
            height = pixbuf.get_height()

            scale_width = max_width
            scale_height = int(height * max_width / width)

            if scale_height > max_height:
                scale_height = max_height
                scale_width = int(width * max_height / height)

            if width > max_width or height > max_height:
                scaled_pixbuf = pixbuf.scale_simple(scale_width, scale_height, GdkPixbuf.InterpType.BILINEAR)
            else:
                scaled_pixbuf = pixbuf

            picture.set_pixbuf(scaled_pixbuf)
        except Exception as e:
            print(f"Error setting image pixbuf: {e}")
            self._on_preset_image_error(picture)

        return False

    def _on_preset_image_error(self, picture: Gtk.Picture) -> bool:
        error_icon = Gtk.Image.new_from_icon_name("image-missing-symbolic")
        error_icon.set_pixel_size(24)
        error_icon.set_halign(Gtk.Align.CENTER)
        error_icon.set_valign(Gtk.Align.CENTER)
        picture.get_parent().set_child(error_icon)
        return False



# Copyright (C) 2025 Alexander Vanhee
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Gtk, Gio, GObject, GLib, Pango
from typing import Optional

class FontDropdown(Gtk.Box):
    __gtype_name__ = "GradiaFontDropdown"
    __gsignals__ = {
        "font-changed": (GObject.SignalFlags.RUN_FIRST, None, (str,))
    }
    selected_font = GObject.Property(
        type=str,
        default="",
        flags=GObject.ParamFlags.READWRITE
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.set_orientation(Gtk.Orientation.VERTICAL)
        self.set_spacing(6)
        self.fonts = self._get_available_fonts()
        self.font_string_list = Gtk.StringList()
        self.font_dropdown = None
        self._updating = False
        self._selected_font_value = ""
        self._setup_font_dropdown()
        self._create_widget()

    def _get_available_fonts(self) -> list[str]:
        widget = Gtk.Label()
        context = widget.get_pango_context()
        font_map = context.get_font_map()
        families = font_map.list_families()
        system_fonts = []
        for family in families:
            font_name = family.get_name()
            system_fonts.append(font_name)
        return sorted(system_fonts)

    def _setup_font_dropdown(self) -> None:
        for font in self.fonts:
            self.font_string_list.append(font)

    def _create_widget(self) -> None:
        self.font_dropdown = Gtk.DropDown(model=self.font_string_list, vexpand=True)
        list_factory = Gtk.SignalListItemFactory()
        list_factory.connect("setup", self._factory_setup_list)
        list_factory.connect("bind", self._factory_bind)
        self.font_dropdown.set_list_factory(list_factory)
        selected_factory = Gtk.SignalListItemFactory()
        selected_factory.connect("setup", self._factory_setup_selected)
        selected_factory.connect("bind", self._factory_bind)
        self.font_dropdown.set_factory(selected_factory)
        self.font_dropdown.connect("notify::selected", self._on_font_selected)
        self.append(self.font_dropdown)

    def _factory_setup_list(self, factory: Gtk.SignalListItemFactory, list_item, *args) -> None:
        label = Gtk.Label(halign=Gtk.Align.START, ellipsize=Pango.EllipsizeMode.NONE)
        label.set_xalign(0)
        list_item.set_child(label)

    def _factory_setup_selected(self, factory: Gtk.SignalListItemFactory, list_item, *args) -> None:
        label = Gtk.Label(halign=Gtk.Align.START, ellipsize=Pango.EllipsizeMode.END, max_width_chars=20)
        label.set_xalign(0)
        list_item.set_child(label)

    def _factory_bind(self, factory: Gtk.SignalListItemFactory, list_item, *args) -> None:
        label = list_item.get_child()
        string_object = list_item.get_item()
        font_name = string_object.get_string()
        label.set_text(font_name)
        attr_list = Pango.AttrList()
        font_desc = Pango.FontDescription.from_string(f"{font_name} 12")
        attr_list.insert(Pango.attr_font_desc_new(font_desc))
        label.set_attributes(attr_list)

    def _on_font_selected(self, dropdown: Gtk.DropDown, _param: GObject.ParamSpec, *args) -> None:
        if self._updating:
            return
        selected_index = dropdown.get_selected()
        if 0 <= selected_index < len(self.fonts):
            font_name = self.fonts[selected_index]
            self._selected_font_value = font_name
            self.emit("font-changed", font_name)

    def do_get_property(self, pspec):
        if pspec.name == "selected-font":
            return self._selected_font_value

    def get_selected_font(self) -> Optional[str]:
        return self._selected_font_value if self._selected_font_value else None

    def set_selected_font(self, font_name: str) -> None:
        if font_name in self.fonts and self._selected_font_value != font_name:
            self._selected_font_value = font_name
            if self.font_dropdown:
                font_index = self.fonts.index(font_name)
                self._updating = True
                self.font_dropdown.set_selected(font_index)
                self._updating = False
            self.notify("selected-font")



# Copyright (C) 2025 Alexander Vanhee
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import cairo
import os
import datetime

import gi
gi.require_version("GtkSource", "5")

from gi.repository import Gtk, Adw, GtkSource, GLib, Gdk, GObject
from gradia.constants import rootdir  # pyright: ignore
from gradia.backend.logger import Logger
from gradia.backend.settings import Settings
from gradia.utils.timestamp_filename import TimestampedFilenameGenerator

logger = Logger()

DEFAULT_WINDOW_WIDTH = 600
MIN_WINDOW_WIDTH = 150
MAX_WINDOW_WIDTH = 1200

class SourceExporter:
    def __init__(self, widget_to_export: Gtk.Widget, padding: int = 24, scale: float = 2.0):
        self.widget = widget_to_export
        self.padding = padding
        self.scale = scale

    def export_to_png(self, out_path: str):
        width = self.widget.get_allocated_width()
        height = self.widget.get_allocated_height()

        total_width = int((width + self.padding * 2) * self.scale)
        total_height = int((height + self.padding * 2) * self.scale)

        surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, total_width, total_height)
        cr = cairo.Context(surface)

        self._setup_transparent_background(cr)


        if not self.widget.get_realized():
            self.widget.realize()

        snapshot = Gtk.Snapshot()
        type(self.widget).do_snapshot(self.widget, snapshot)
        render_node = snapshot.to_node()

        if render_node:
            cr.save()
            cr.scale(self.scale, self.scale)
            cr.translate(self.padding, self.padding)

            self._apply_rounded_clipping(cr, width, height)

            render_node.draw(cr)
            cr.restore()

        surface.write_to_png(out_path)
        logger.info(f"Exported source image to: {out_path}")

    def _setup_transparent_background(self, cr):
        cr.set_operator(cairo.OPERATOR_CLEAR)
        cr.paint()
        cr.set_operator(cairo.OPERATOR_OVER)

    def _apply_rounded_clipping(self, cr, width, height, radius=12):
        cr.new_path()
        cr.arc(radius, radius, radius, 3.14159, 3 * 3.14159 / 2)
        cr.arc(width - radius, radius, radius, 3 * 3.14159 / 2, 0)
        cr.arc(width - radius, height - radius, radius, 0, 3.14159 / 2)
        cr.arc(radius, height - radius, radius, 3.14159 / 2, 3.14159)
        cr.close_path()
        cr.clip()


class ResizeHandle(Gtk.Box):
    def __init__(self):
        super().__init__()
        self.set_size_request(10, -1)
        self.set_cursor(Gdk.Cursor.new_from_name("ew-resize"))
        self.get_style_context().add_class("resize-handle")

        separator = Gtk.Separator(orientation=Gtk.Orientation.VERTICAL)
        separator.set_valign(Gtk.Align.FILL)
        separator.set_opacity(0.0)
        self.append(separator)


class DragController:
    def __init__(self, content_box, resize_handle):
        self.content_box = content_box
        self.resize_handle = resize_handle
        self.initial_width = DEFAULT_WINDOW_WIDTH
        self.dragging = False

        self._setup_controllers()

    def _setup_controllers(self):
        self.motion_controller = Gtk.EventControllerMotion()
        self.resize_handle.add_controller(self.motion_controller)
        self.motion_controller.connect("motion", self._on_motion)

        self.drag_gesture = Gtk.GestureDrag()
        self.drag_gesture.set_button(1)
        self.resize_handle.add_controller(self.drag_gesture)

        self.drag_gesture.connect("drag-begin", self._on_drag_begin)
        self.drag_gesture.connect("drag-update", self._on_drag_update)
        self.drag_gesture.connect("drag-end", self._on_drag_end)

    def _on_drag_begin(self, gesture, start_x, start_y):
        self.initial_width = self.content_box.get_allocated_width()
        self.dragging = True

    def _on_drag_update(self, gesture, offset_x, offset_y):
        if not self.dragging:
            return

        new_width = max(MIN_WINDOW_WIDTH,
                        min(MAX_WINDOW_WIDTH, int(self.initial_width + offset_x)))

        self.content_box.set_size_request(new_width, -1)
        self.content_box.queue_allocate()

    def _on_drag_end(self, gesture, offset_x, offset_y):
        self.dragging = False

    def _on_motion(self, controller, x, y):
        if self.dragging:
            self.content_box.queue_draw()


class ResizableContainer(Gtk.Box):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.set_orientation(Gtk.Orientation.HORIZONTAL)
        self.set_halign(Gtk.Align.CENTER)
        self.set_valign(Gtk.Align.START)

        self.content_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self.content_box.set_size_request(DEFAULT_WINDOW_WIDTH, -1)

        self.resize_handle = ResizeHandle()
        self.drag_controller = DragController(self.content_box, self.resize_handle)

        self.append(self.content_box)
        self.append(self.resize_handle)

    def set_child_widget(self, widget):
        child = self.content_box.get_first_child()
        if child:
            self.content_box.remove(child)

        if widget:
            self.content_box.append(widget)

    def get_content_width(self):
        return self.content_box.get_allocated_width()


class SourceViewManager:
    def __init__(self):
        self.source_view = GtkSource.View.new()
        self.source_buffer = self.source_view.get_buffer()
        self.source_buffer.set_highlight_matching_brackets(False)
        self._text_changed_callback = None
        self._setup_source_view()
        self._connect_signals()

    def _setup_source_view(self):
        self.source_view.set_top_margin(10)
        self.source_view.set_right_margin(10)
        self.source_view.set_left_margin(10)
        self.source_view.set_bottom_margin(10)
        self.source_view.set_valign(Gtk.Align.START)
        self.source_buffer.set_highlight_syntax(True)
        self.source_view.set_monospace(True)
        self.source_view.set_wrap_mode(Gtk.WrapMode.WORD_CHAR)
        self.source_view.set_show_line_numbers(True)

    def _connect_signals(self):
        self.source_buffer.connect('changed', self._on_text_changed)

    def _on_text_changed(self, buffer):
        if self._text_changed_callback:
            text = buffer.get_text(buffer.get_start_iter(), buffer.get_end_iter(), False)
            self._text_changed_callback(text)

    def set_text(self, text):
        self.source_buffer.set_text(text)

    def get_text(self):
        return self.source_buffer.get_text(
            self.source_buffer.get_start_iter(),
            self.source_buffer.get_end_iter(),
            False
        )

    def set_language(self, language):
        self.source_buffer.set_language(language)

    def set_style_scheme(self, scheme):
        self.source_buffer.set_style_scheme(scheme)

    def set_show_line_numbers(self, show_numbers: bool):
        self.source_view.set_show_line_numbers(show_numbers)

    def set_text_changed_callback(self, callback):
        self._text_changed_callback = callback

    def get_view(self):
        return self.source_view

class LanguageManager:
    def __init__(self):
        self.lang_manager = GtkSource.LanguageManager.get_default()
        self.languages = sorted(self.lang_manager.get_language_ids())

    def get_languages(self):
        return self.languages

    def get_language(self, lang_id):
        return self.lang_manager.get_language(lang_id)


class StyleManager:
    def __init__(self):
        self.style_manager = GtkSource.StyleSchemeManager.get_default()
        self.style_manager.append_search_path("resource:///be/alexandervanhee/gradia/source-styles")
        self.settings = Gtk.Settings.get_default()
        self.theme_changed_callbacks = []

        self.style_schemes = [
            'Adwaita',
            'classic',
            'cobalt-light',
            'kate',
            'solarized-light',
            'tango',
            'Adwaita-dark',
            'classic-dark',
            'cobalt',
            'kate-dark',
            'oblivion',
            'solarized-dark',
            'clone-of-ubuntu',
            'builder-dark',
            'vscode-dark'
        ]

        self.use_generic_styles = True

    def get_all_schemes(self):
        return self.style_schemes


class FakeWindowManager:
    def __init__(self, source_view):
        self.source_view = source_view
        self.fake_window_container = None
        self.header_bar = None
        self.title_entry = None
        self.current_style_provider = None
        self.settings = Settings()

    def create_fake_window(self):
        if self.fake_window_container:
            return self.fake_window_container

        frame = Gtk.Frame(valign=Gtk.Align.START, margin_top=12)
        frame.add_css_class("window-border")
        frame.add_css_class("card")

        self.fake_window_container = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=0)
        self.fake_window_container.get_style_context().add_class("adw-window")

        self.header_bar = Adw.HeaderBar.new()
        self.header_bar.get_style_context().add_class("flat")

        self.title_entry = Gtk.Entry(xalign=0.5, focus_on_click=False)
        self.title_entry.set_text(self.settings.source_snippet_title)
        self.title_entry.set_halign(Gtk.Align.CENTER)
        self.title_entry.set_valign(Gtk.Align.CENTER)
        self.title_entry.set_width_chars(45)
        self.title_entry.set_max_length(100)
        self.title_entry.set_has_frame(False)
        self.title_entry.get_style_context().add_class("title")
        self.title_entry.get_style_context().add_class("title-entry")

        def on_title_entry_changed(entry):
            new_title = entry.get_text()
            self.settings.source_snippet_title = new_title

        self.title_entry.connect("changed", on_title_entry_changed)

        self.header_bar.set_title_widget(self.title_entry)

        source_view_container = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        source_view_container.append(self.source_view)

        self.fake_window_container.append(self.header_bar)
        self.fake_window_container.append(source_view_container)

        frame.set_child(self.fake_window_container)

        return frame

    def update_header_colors(self, style_scheme):
        if not self.header_bar or not self.title_entry:
            return

        if self.current_style_provider:
            self.header_bar.get_style_context().remove_provider(self.current_style_provider)
            self.title_entry.get_style_context().remove_provider(self.current_style_provider)

        bg_color, fg_color = self._extract_header_colors(style_scheme)

        if bg_color and fg_color:
            css_data = f"""
            headerbar {{
                background: {bg_color};
                color: {fg_color};
                border-bottom: 1px solid alpha({fg_color}, 0.1);
            }}

            """

            self.current_style_provider = Gtk.CssProvider()
            self.current_style_provider.load_from_data(css_data.encode())

            self.header_bar.get_style_context().add_provider(
                self.current_style_provider,
                Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
            )
            self.title_entry.get_style_context().add_provider(
                self.current_style_provider,
                Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
            )

    def _extract_header_colors(self, style_scheme):
        scheme_id = style_scheme.get_id()

        text_style = style_scheme.get_style("text")
        bg_color = text_style.get_property("background")
        fg_color = text_style.get_property("foreground")

        return bg_color, fg_color


    def destroy_fake_window(self):
        if self.current_style_provider and self.header_bar:
            self.header_bar.get_style_context().remove_provider(self.current_style_provider)
            if self.title_entry:
                self.title_entry.get_style_context().remove_provider(self.current_style_provider)

        self.fake_window_container = None
        self.header_bar = None
        self.title_entry = None
        self.current_style_provider = None

    def get_container(self):
        return self.fake_window_container


@Gtk.Template(resource_path=f"{rootdir}/ui/image_generator_window.ui")
class SourceImageGeneratorWindow(Adw.Window):
    __gtype_name__ = "SourceviewtestWindow"

    export_button = Gtk.Template.Child()
    scroller = Gtk.Template.Child()
    language_dropdown = Gtk.Template.Child()
    style_scheme_button = Gtk.Template.Child()
    style_scheme_popover = Gtk.Template.Child()
    style_scheme_flowbox = Gtk.Template.Child()
    style_scheme_label = Gtk.Template.Child()
    fake_window_button = Gtk.Template.Child()
    line_numbers_button = Gtk.Template.Child()
    toolbar_view = Gtk.Template.Child()

    def __init__(self, parent_window, temp_dir=None, export_callback=None, **kwargs):
        super().__init__(
            title=_("Source Snippets"),
            modal=True,
            transient_for=parent_window,
            **kwargs
        )
        self.settings = Settings()

        self.temp_dir = temp_dir
        self.export_callback = export_callback
        self.parent_window = parent_window

        self.source_view_manager = SourceViewManager()
        self.language_manager = LanguageManager()
        self.style_manager = StyleManager()
        self.fake_window_manager = FakeWindowManager(self.source_view_manager.get_view())

        self.current_scheme_id = self.settings.source_snippet_style_scheme
        self.gtk_style_manager = GtkSource.StyleSchemeManager.get_default()
        self.style_scheme_previews = {}

        self._setup_ui()
        self._setup_dropdowns()
        self._setup_initial_state()
        self._connect_signals()

        self.settings.bind_switch(self.fake_window_button, "source-snippet-show-frame")
        self.settings.bind_switch(self.line_numbers_button, "source-snippet-show-line-numbers")

        shortcut_controller = Gtk.ShortcutController()
        shortcut = Gtk.Shortcut.new(
            Gtk.ShortcutTrigger.parse_string("Escape"),
            Gtk.ShortcutAction.parse_string("action(window.close)")
        )
        shortcut_controller.add_shortcut(shortcut)
        self.add_controller(shortcut_controller)


    def _setup_ui(self):
        self.resizable_container = ResizableContainer()
        self.scroller.set_child(self.resizable_container)
        self.source_view_manager.set_text(self.settings.source_snippet_code_text)

        if self.current_scheme_id:
            scheme = self.gtk_style_manager.get_scheme(self.current_scheme_id)
            if scheme:
                self.source_view_manager.set_style_scheme(scheme)

        def update_settings(text):
            self.settings.source_snippet_code_text = text

        self.source_view_manager.set_text_changed_callback(update_settings)

    def _setup_dropdowns(self):
        languages = self.language_manager.get_languages()
        self.language_dropdown.set_model(Gtk.StringList.new(languages))

        expression = Gtk.ClosureExpression.new(
            GObject.TYPE_STRING, lambda obj: obj.get_string(), None
        )

        self.language_dropdown.set_expression(expression)
        initial_language = self.settings.source_snippet_language

        if initial_language in languages:
            index = languages.index(initial_language)
            self.language_dropdown.set_selected(index)
            language = self.language_manager.get_language(initial_language)
            self.source_view_manager.set_language(language)

        self._setup_style_scheme_flowbox()

    def _setup_style_scheme_flowbox(self):
        scheme_manager = GtkSource.StyleSchemeManager.get_default()
        scheme_ids = self.style_manager.get_all_schemes()
        self.style_scheme_previews = {}

        for scheme_id in scheme_ids:
            scheme = scheme_manager.get_scheme(scheme_id)
            if not scheme:
                continue
            preview = GtkSource.StyleSchemePreview.new(scheme)
            preview.set_valign(Gtk.Align.START)
            preview.set_hexpand(True)
            preview.set_vexpand(False)
            flowboxchild = Gtk.FlowBoxChild(valign=Gtk.Align.START)
            flowboxchild.set_child(preview)
            self.style_scheme_flowbox.append(flowboxchild)
            self.style_scheme_previews[scheme_id] = preview
            if scheme_id == self.current_scheme_id:
                preview.set_selected(True)
                self.style_scheme_label.set_text(scheme.get_name())
            preview.connect("activate", self._on_style_scheme_selected, scheme_id, scheme.get_name())

            GLib.idle_add(self._force_popover_sizing)

    def _force_popover_sizing(self):
        self.style_scheme_popover.set_opacity(0)
        self.style_scheme_popover.popup()
        GLib.idle_add(self._hide_popover_after_sizing)
        return False

    def _hide_popover_after_sizing(self):
        self.style_scheme_popover.set_opacity(1)
        self.style_scheme_popover.popdown()
        return False

    def _setup_initial_state(self):
        self.fake_window_button.set_active(True)
        self.line_numbers_button.set_active(True)
        self._update_view_mode()
        self._update_line_numbers()

    def _connect_signals(self):
        self.export_button.connect("clicked", self._on_export_clicked)
        self.language_dropdown.connect("notify::selected", self._on_language_changed)
        self.fake_window_button.connect("notify::state", self._on_fake_window_toggled)
        self.line_numbers_button.connect("notify::state", self._on_line_numbers_toggled)

    def _safely_unparent_widget(self, widget):
        parent = widget.get_parent()
        if parent:
            if hasattr(parent, 'remove'):
                parent.remove(widget)
            elif hasattr(parent, 'set_child'):
                parent.set_child(None)

    def _update_view_mode(self):
        self.resizable_container.set_child_widget(None)

        source_view = self.source_view_manager.get_view()
        self._safely_unparent_widget(source_view)

        if self.fake_window_button.get_active():
            fake_window_frame = self.fake_window_manager.create_fake_window()
            self.resizable_container.set_child_widget(fake_window_frame)

            current_scheme = self._get_current_scheme()
            if current_scheme:
                self.fake_window_manager.update_header_colors(current_scheme)
        else:
            frame = Gtk.Frame(valign=Gtk.Align.START, margin_top=12)
            frame.add_css_class("window-border")
            frame.add_css_class("card")
            frame.set_child(source_view)
            self.resizable_container.set_child_widget(frame)

    def _update_line_numbers(self):
        show_line_numbers = self.line_numbers_button.get_active()
        self.source_view_manager.set_show_line_numbers(show_line_numbers)

    def _get_current_scheme(self):
        if self.current_scheme_id:
            return self.gtk_style_manager.get_scheme(self.current_scheme_id)
        return None

    def _on_export_clicked(self, _button):
        timestamp = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
        filename = TimestampedFilenameGenerator().generate(_("Source Snippet From %Y-%m-%d %H-%M-%S")) + ".png"

        if self.temp_dir:
            output_path = os.path.join(self.temp_dir, filename)
        else:
            output_path = os.path.join(os.getcwd(), filename)

        widget_to_export = self._get_export_widget()
        widget_to_export.get_root().grab_focus()

        self.source_view_manager.get_view().get_style_context().add_class("no-highlights")
        if self.fake_window_manager.title_entry is not None:
            self.fake_window_manager.title_entry.get_style_context().add_class("no-highlights")

        def do_export():
            exporter = SourceExporter(widget_to_export)
            exporter.export_to_png(output_path)
            self.export_callback(output_path)
            self.close()
            return False

        GLib.timeout_add(50, do_export)

    def _on_language_changed(self, dropdown, _param):
        languages = self.language_manager.get_languages()
        index = dropdown.get_selected()
        if 0 <= index < len(languages):
            language = self.language_manager.get_language(languages[index])
            self.source_view_manager.set_language(language)
            self.settings.source_snippet_language = languages[index]

    def _on_style_scheme_selected(self, button, scheme_id, scheme_name):
        self.current_scheme_id = scheme_id
        self.settings.source_snippet_style_scheme = scheme_id

        for sid, preview in self.style_scheme_previews.items():
            preview.set_selected(sid == scheme_id)

        scheme = self.gtk_style_manager.get_scheme(scheme_id)
        if scheme:
            self.source_view_manager.set_style_scheme(scheme)
            if self.fake_window_button.get_active():
                self.fake_window_manager.update_header_colors(scheme)

        self.style_scheme_label.set_text(scheme_name)
        self.style_scheme_popover.popdown()

    def _on_fake_window_toggled(self, switch, param_spec):
        if not switch.get_state():
            source_view = self.source_view_manager.get_view()
            self._safely_unparent_widget(source_view)
            self.fake_window_manager.destroy_fake_window()
        self._update_view_mode()

    def _on_line_numbers_toggled(self, switch, param_spec):
        self._update_line_numbers()

    def _get_export_widget(self) -> Gtk.Widget:
        if self.fake_window_button.get_active():
            return self.fake_window_manager.get_container()
        else:
            return self.source_view_manager.get_view()



# Copyright (C) 2025 Alexander Vanhee
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Adw, Gtk, GLib, Gdk
from typing import Callable, Optional
from gradia.graphics.gradient import Gradient
from gradia.constants import rootdir
from gradia.utils.colors import parse_rgb_string


class GradientStepDialogButton(Gtk.Button):
    __gtype_name__ = "GradiaGradientStepDialogButton"

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

        self._gradient = None
        self._callback = None
        self._dialog = None

        self.set_icon_name("edit-symbolic")
        self.set_valign(Gtk.Align.CENTER)
        self.add_css_class("flat")
        self.connect("clicked", self._on_button_clicked)

    def _on_button_clicked(self, button: Gtk.Button) -> None:
        if self._gradient is None:
            raise ValueError("No gradient set. Use set_gradient() before showing the dialog.")

        if self._dialog is None:
            self._dialog = CssGradientDialog()
            self._dialog.set_callback(self._on_dialog_apply)

        self._dialog.set_gradient(self._gradient)
        self._dialog.present(self.get_root())

    def _on_dialog_apply(self, gradient: Gradient) -> None:
        self._gradient = gradient
        if self._callback:
            self._callback(gradient)

    def set_gradient(self, gradient: Gradient) -> None:
        self._gradient = gradient

    def get_gradient(self) -> Gradient:
        return self._gradient

    def set_callback(self, callback: Callable[[Gradient], None]) -> None:
        self._callback = callback


class GradientStepRow(Adw.ActionRow):
    def __init__(self, step_index: int, position: float, color: str, on_changed_callback: Callable):
        super().__init__()

        self.step_index = step_index
        self.on_changed_callback = on_changed_callback

        self.color_button = Gtk.ColorButton()
        self.color_button.set_valign(Gtk.Align.CENTER)
        self._set_color_from_string(color)
        self.color_button.connect("color-set", self._on_color_changed)
        self.add_prefix(self.color_button)

        position_adjustment = Gtk.Adjustment(value=position * 100, lower=0, upper=100, step_increment=1, page_increment=10)
        self.position_spinbutton = Gtk.SpinButton(adjustment=position_adjustment, digits=1)
        self.position_spinbutton.set_value(position * 100)
        self.position_spinbutton.connect("value-changed", self._on_position_changed)


        position_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=0)
        position_box.append(self.position_spinbutton)

        self.remove_button = Gtk.Button(icon_name="user-trash-symbolic")
        self.remove_button.add_css_class("destructive-action")
        self.remove_button.connect("clicked", self._on_remove_clicked)

        controls_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=6 , valign=Gtk.Align.CENTER)
        controls_box.append(position_box)

        controls_box.append(self.remove_button)

        self.add_suffix(controls_box)

    def _set_color_from_string(self, color_str: str):
        rgb = parse_rgb_string(color_str)
        if rgb:
            r, g, b = rgb
            rgba = Gdk.RGBA()
            rgba.red = r / 255.0
            rgba.green = g / 255.0
            rgba.blue = b / 255.0
            rgba.alpha = 1.0
            self.color_button.set_rgba(rgba)

    def _on_position_changed(self, spinbutton):
        self.on_changed_callback()

    def _on_color_changed(self, color_button):
        self.on_changed_callback()

    def _on_remove_clicked(self, button):
        self.on_changed_callback(remove_index=self.step_index)

    def get_position(self) -> float:
        return self.position_spinbutton.get_value() / 100.0

    def get_color(self) -> str:
        rgba = self.color_button.get_rgba()
        r = int(rgba.red * 255)
        g = int(rgba.green * 255)
        b = int(rgba.blue * 255)
        return f"rgb({r},{g},{b})"

    def update_remove_button_sensitivity(self, can_remove: bool):
        self.remove_button.set_sensitive(can_remove)


@Gtk.Template(resource_path=f"{rootdir}/ui/gradient_step_dialog.ui")
class CssGradientDialog(Adw.Dialog):
    __gtype_name__ = "GradiaGradientStepDialog"

    preview_box = Gtk.Template.Child()
    steps_group = Gtk.Template.Child()
    add_step_button = Gtk.Template.Child()

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

        self._gradient = None
        self._callback = None
        self._initial_gradient = None
        self._step_rows = []

        self.add_step_button.connect("clicked", self._on_add_step_clicked)
        self.connect("closed", self._on_dialog_closed)

    def set_gradient(self, gradient: Gradient) -> None:
        self._gradient = gradient
        self._initial_gradient = gradient
        self._populate_steps()
        self._update_preview()

    def set_callback(self, callback: Callable[[Gradient], None]) -> None:
        self._callback = callback

    def _populate_steps(self):
        for row in self._step_rows:
            self.steps_group.remove(row)
        self._step_rows.clear()

        for i, (position, color) in enumerate(self._gradient.steps):
            row = GradientStepRow(i, position, color, self._on_step_changed)
            self._step_rows.append(row)
            self.steps_group.add(row)

        self._update_remove_button_sensitivity()
        self._update_add_button_visibility()

    def _update_remove_button_sensitivity(self):
        can_remove = len(self._step_rows) > 2
        for row in self._step_rows:
            row.update_remove_button_sensitivity(can_remove)

    def _on_step_changed(self, remove_index=None):
        if remove_index is not None:
            if len(self._step_rows) > 2:
                row_to_remove = self._step_rows[remove_index]
                self.steps_group.remove(row_to_remove)
                self._step_rows.remove(row_to_remove)

        steps = []
        for row in self._step_rows:
            position = row.get_position()
            color = row.get_color()
            steps.append((position, color))

        steps.sort(key=lambda x: x[0])

        self._gradient = Gradient(
            mode=self._gradient.mode,
            steps=steps,
            angle=self._gradient.angle
        )

        self._update_preview()
        self._update_remove_button_sensitivity()
        self._update_add_button_visibility()

    def _on_add_step_clicked(self, button):
        if len(self._step_rows) < 5:
            new_position = 0.5
            new_color = "rgb(128,128,128)"

            if len(self._step_rows) >= 2:
                positions = [row.get_position() for row in self._step_rows]
                positions.sort()
                for i in range(len(positions) - 1):
                    gap = positions[i + 1] - positions[i]
                    if gap > 0.1:
                        new_position = positions[i] + gap / 2
                        break

            row = GradientStepRow(len(self._step_rows), new_position, new_color, self._on_step_changed)
            self._step_rows.append(row)
            self.steps_group.add(row)

            self._on_step_changed()

    def _update_add_button_visibility(self):
        self.add_step_button.set_sensitive(len(self._step_rows) < 5)

    def _update_preview(self) -> None:
        if self._gradient is not None:
            self._apply_preview_gradient(self._gradient)

    def _apply_preview_gradient(self, gradient: Gradient) -> None:
        css = f"""
            .gradient-preview {{
                background-image: {gradient.to_css()};
            }}
        """

        css_provider = Gtk.CssProvider()
        css_provider.load_from_string(css)

        style_context = self.preview_box.get_style_context()

        style_context.add_provider(
            css_provider,
            Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION + 5
        )

    def _on_dialog_closed(self, dialog: Adw.Dialog) -> None:
        if self._gradient and self._callback:
            if not self._initial_gradient or self._gradient.to_css() != self._initial_gradient.to_css():
                self._ensure_minimum_spacing()
                self._callback(self._gradient)

    def _ensure_minimum_spacing(self, min_gap: float = 0.18):
        steps = sorted(self._gradient.steps, key=lambda x: x[0])

        if len(steps) < 2:
            return

        adjusted = [steps[0]]

        for i in range(1, len(steps)):
            pos, color = steps[i]
            prev_pos = adjusted[-1][0]

            if pos - prev_pos < min_gap:
                pos = prev_pos + min_gap

            if pos > 1.0:
                pos = 1.0

            adjusted.append((pos, color))

        for i in range(len(adjusted) - 2, -1, -1):
            pos, color = adjusted[i]
            next_pos = adjusted[i + 1][0]

            if next_pos - pos < min_gap:
                pos = max(0.0, next_pos - min_gap)
                adjusted[i] = (pos, color)

        self._gradient = Gradient(
            mode=self._gradient.mode,
            steps=adjusted,
            angle=self._gradient.angle
        )



# Copyright (C) 2025 Alexander Vanhee
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import os
from gi.repository import Gdk, GLib, GdkPixbuf

def save_texture_to_file(texture, temp_dir: str) -> str:
    temp_path: str = os.path.join(temp_dir, f"clipboard_image_{os.urandom(6).hex()}.png")
    texture.save_to_png(temp_path)
    return temp_path

def save_pixbuff_to_path(temp_dir: str, pixbuff: GdkPixbuf.Pixbuf) -> str:
    TEMP_FILE_NAME: str = "clipboard_temp.png"
    temp_path: str = os.path.join(temp_dir, TEMP_FILE_NAME)
    pixbuff.savev(temp_path, "png", [], [])
    return temp_path


def copy_text_to_clipboard(text: str) -> None:
    display = Gdk.Display.get_default()
    if not display:
        print("Warning: Failed to retrieve `Gdk.Display` object.")
        return

    clipboard: Gdk.Clipboard = display.get_clipboard()
    text_bytes = text.encode("utf-8")
    bytes_data = GLib.Bytes.new(text_bytes)
    content_provider = Gdk.ContentProvider.new_for_bytes("text/plain;charset=utf-8", bytes_data)
    clipboard.set_content(content_provider)

def copy_pixbuf_to_clipboard(pixbuf: GdkPixbuf.Pixbuf) -> None:
    display = Gdk.Display.get_default()
    if not display:
        print("Warning: Failed to retrieve `Gdk.Display` object.")
        return

    clipboard: Gdk.Clipboard = display.get_clipboard()
    content_provider: Gdk.ContentProvider = Gdk.ContentProvider.new_for_value(pixbuf)
    clipboard.set_content(content_provider)



# Copyright (C) 2025 Alexander Vanhee, tfuxu
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import io
import os
import tempfile
import threading
import shutil
from pathlib import Path
from typing import Callable, Optional

from PIL import Image
from gi.repository import Adw, GLib, Gdk, GdkPixbuf, Gio, Gtk

from gradia.constants import rootdir
from gradia.graphics.background import Background
from gradia.ui.widget.preset_button import ImagePresetButton
from gradia.app_constants import PRESET_IMAGES
from gradia.app_constants import SUPPORTED_EXPORT_FORMATS

class ImageBackground(Background):
    @property
    def SAVED_IMAGE_PATH(self) -> Path:
        cache_dir = GLib.get_user_cache_dir()
        return Path(cache_dir) / 'gradia' / 'last_image.png'

    def __init__(self, file_path: Optional[str] = None) -> None:
        self.file_path: Optional[str] = file_path
        self.image: Optional[Image.Image] = None

        self.SAVED_IMAGE_PATH.parent.mkdir(parents=True, exist_ok=True)

        if file_path:
            self.load_image(file_path)
        else:
            if self.SAVED_IMAGE_PATH.exists():
                self.load_image(str(self.SAVED_IMAGE_PATH))
            elif PRESET_IMAGES:
                self.load_image(PRESET_IMAGES[0])

    def load_image(self, path: str) -> None:
        self.file_path = path

        if path.startswith(rootdir):
            resource_data = Gio.resources_lookup_data(path, Gio.ResourceLookupFlags.NONE)
            bytes_data = resource_data.get_data()

            if bytes_data is None:
                raise RuntimeError("Failed to get data from resource lookup")

            byte_stream = io.BytesIO(bytes_data)
            self.image = Image.open(byte_stream).convert("RGBA")
        else:
            self.image = Image.open(path).convert("RGBA")

    def save_image_copy_async(self) -> None:
        def save_in_background():
            if self.image:
                _, extension = os.path.splitext(self.SAVED_IMAGE_PATH)
                target_format = 'PNG'
                for format in SUPPORTED_EXPORT_FORMATS:
                    if extension.lower() in format['extensions']:
                        target_format = format['shortname']
                        break
                self.image.save(self.SAVED_IMAGE_PATH, target_format)

        thread = threading.Thread(target=save_in_background, daemon=True)
        thread.start()

    def prepare_image(self, width: int, height: int) -> Optional[Image.Image]:
        if not self.image:
            return None

        img = self.image
        img_ratio = img.width / img.height
        target_ratio = width / height

        if img_ratio > target_ratio:
            new_height = height
            new_width = int(new_height * img_ratio)
        else:
            new_width = width
            new_height = int(new_width / img_ratio)

        img_resized = img.resize((new_width, new_height), Image.Resampling.LANCZOS)
        left = (new_width - width) // 2
        top = (new_height - height) // 2
        right = left + width
        bottom = top + height

        img_cropped = img_resized.crop((left, top, right, bottom))
        return img_cropped

    def get_name(self) -> str:
        return f"image-{self.file_path or 'none'}"


@Gtk.Template(resource_path=f"{rootdir}/ui/selectors/image_selector.ui")
class ImageSelector(Adw.PreferencesGroup):
    __gtype_name__ = "GradiaImageSelector"

    preset_button: ImagePresetButton = Gtk.Template.Child()
    preview_picture: Gtk.Picture = Gtk.Template.Child()
    open_image_dialog: Gtk.FileDialog = Gtk.Template.Child()
    image_filter: Gtk.FileFilter = Gtk.Template.Child()

    def __init__(
        self,
        image_background: ImageBackground,
        callback: Optional[Callable[[ImageBackground], None]] = None,
        **kwargs
    ) -> None:
        super().__init__(**kwargs)

        self.image_background = image_background
        self.callback = callback

        self._setup_file_dialog()
        self._setup_drag_and_drop()
        self._setup_gesture()
        self._setup_preset_button()

        self._update_preview()

    def _setup_file_dialog(self) -> None:
        filter_list = Gio.ListStore.new(Gtk.FileFilter)
        filter_list.append(self.image_filter)
        self.open_image_dialog.set_filters(filter_list)

    def _setup_drag_and_drop(self) -> None:
        drop_target = Gtk.DropTarget.new(Gio.File, Gdk.DragAction.COPY)
        drop_target.connect("drop", self._on_image_drop)
        self.preview_picture.add_controller(drop_target)

    def _setup_gesture(self) -> None:
        gesture = Gtk.GestureClick.new()
        gesture.connect("pressed", self._on_preview_clicked)
        self.preview_picture.add_controller(gesture)

    def _setup_preset_button(self) -> None:
        self.preset_button.set_callback(self._on_preset_selected)

    @Gtk.Template.Callback()
    def _on_select_clicked(self, _button: Gtk.Button, *args) -> None:
        self.open_image_dialog.open(self.get_root(), None, self._on_file_dialog_ready)

    def _on_preset_selected(self, path: str) -> None:
        self._load_image_async(path)

    def _on_preview_clicked(self, _gesture: Gtk.GestureClick, _n_press: int, _x: float, _y: float) -> None:
        self.open_image_dialog.open(self.get_root(), None, self._on_file_dialog_ready)

    def _on_image_drop(self, _drop_target: Gtk.DropTarget, file: Gio.File, _x: int, _y: int) -> bool:
        file_path = file.get_path()
        if file_path:
            self._load_image_async(file_path)
            return True
        return False

    def _on_file_dialog_ready(self, file_dialog: Gtk.FileDialog, result: Gio.AsyncResult) -> None:
        try:
            file = file_dialog.open_finish(result)
            file_path = file.get_path()
            if file_path:
                self._load_image_async(file_path)
        except GLib.Error as e:
            print(f"FileDialog cancelled or failed: {e.message}")

    def _on_image_loaded(self) -> None:
        self._update_preview()
        self.image_background.save_image_copy_async()
        if self.callback:
            self.callback(self.image_background)

    def _update_preview(self) -> None:
        if self.image_background.image:
            def save_and_update() -> None:
                try:
                    if not self.image_background.image:
                        return

                    image = self.image_background.image.copy()

                    max_width = 400
                    if image.width > max_width:
                        ratio = max_width / image.width
                        new_size = (int(image.width * ratio), int(image.height * ratio))
                        image = image.resize(new_size, Image.LANCZOS)

                    with tempfile.NamedTemporaryFile(suffix='.png', delete=False) as temp_file:
                        temp_path = temp_file.name
                        image.save(temp_path, 'PNG')

                    GLib.idle_add(self._set_preview_image, temp_path)
                except Exception as e:
                    print(f"Error saving preview: {e}")

            thread = threading.Thread(target=save_and_update, daemon=True)
            thread.start()
        else:
            self.preview_picture.set_paintable(None)

    def _set_preview_image(self, temp_path: str) -> bool:
        self.preview_picture.set_filename(temp_path)
        def cleanup_temp_file():
            try:
                os.unlink(temp_path)
            except Exception:
                pass
            return False

        GLib.timeout_add(100, cleanup_temp_file)
        return False

    def _load_image_async(self, file_path: str) -> None:
        def load_in_background():
            try:
                self.image_background.load_image(file_path)
                GLib.idle_add(self._on_image_loaded)
            except Exception as e:
                print(f"Error loading image: {e}")
        thread = threading.Thread(target=load_in_background, daemon=True)
        thread.start()



# Copyright (C) 2025 Alexander Vanhee, tfuxu
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Optional
from gi.repository import Adw, Gtk
from gradia import constants


class AboutDialog:
    def __init__(self, version: str):
        self.version = version
        self.dialog = None

    def create(self) -> Adw.AboutDialog:
        self.dialog = Adw.AboutDialog(
            application_name="Gradia",
            version=self.version,
            comments=_("Make your images ready for all"),
            website="https://github.com/AlexanderVanhee/Gradia",
            issue_url="https://github.com/AlexanderVanhee/Gradia/issues",
            developer_name="Alexander Vanhee",
            developers=[
                "Alexander Vanhee https://github.com/AlexanderVanhee",
                "tfuxu https://github.com/tfuxu",
            ],
            designers=[
                "drpetrikov https://github.com/drpetrikov"
            ],
            application_icon=constants.app_id,
            # Translators: This is a place to put your credits (formats: "Name https://example.com" or "Name <email@example.com>", no quotes) and is not meant to be translated literally.
            translator_credits=_("translator-credits"),
            copyright="Copyright © 2025 Alexander Vanhee",
            license_type=Gtk.License.GPL_3_0
        )

        self.dialog.add_acknowledgement_section(
            _("Code and Design Borrowed from"),
            [
                "Switcheroo https://apps.gnome.org/en-GB/Converter/",
                "Halftone https://github.com/tfuxu/Halftone",
                "Gradience https://github.com/GradienceTeam/Gradience",
                "Emblem https://apps.gnome.org/en-GB/Emblem/",
                "Builder https://apps.gnome.org/en-GB/Builder/",
                None
            ]
        )

        self.dialog.add_acknowledgement_section(
            _("Image Sources"),
            [
                "GNOME backgrounds https://gitlab.gnome.org/GNOME/gnome-backgrounds",
                "Fruit Basket https://unsplash.com/photos/background-pattern-oWr5S1bO2ak",
                "Codioful https://unsplash.com/@codioful",
                None
            ]
        )


        return self.dialog

    def show(self, parent: Optional[Gtk.Window] = None):
        if not self.dialog:
            self.create()
        self.dialog.present(parent)


class ShortcutsDialog:
    def __init__(self, parent: Optional[Gtk.Window] = None):
        self.parent = parent
        self.dialog = None
        self.create()
        self.show()

    def _has_adw_shortcuts_dialog(self):
        try:
            return (hasattr(Adw, 'ShortcutsDialog') and
                    Adw.get_major_version() >= 1 and
                    Adw.get_minor_version() >= 8)
        except (NameError, AttributeError):
            return False

    def create(self):
        sections_data = [
            (_("File Actions"), [
                (_("Open Image"), "<Ctrl>O"),
                (_("Take a Screenshot"), "<Ctrl>A"),
                (_("Save to File"), "<Ctrl>S"),
                (_("Copy Image to Clipboard"), "<Ctrl>C"),
                (_("Paste From Clipboard"), "<Ctrl>V"),
                (_("Share Image"), "<Ctrl>M"),
            ]),
            (_("Annotations"), [
                (_("Undo"), "<Ctrl>Z"),
                (_("Redo"), "<Ctrl><Shift>Z"),
                (_("Erase Selected"), "Delete"),
                (_("Select"), "0 S"),
                (_("Pen"), "1 P"),
                (_("Text"), "2 T"),
                (_("Line"), "3 L"),
                (_("Arrow"), "4 A"),
                (_("Rectangle"), "5 R"),
                (_("Oval"), "6 O"),
                (_("Highlighter"), "7 H"),
                (_("Censor"), "8 C"),
                (_("Number"), "9 N"),
            ]),
            (_("Cropping"), [
                (_("Toggle Crop Mode"), "<Ctrl>R"),
                (_("Reset Crop"), "<Ctrl><Shift>R")
            ]),
            (_("Zooming"), [
                (_("Zoom In"), "<Ctrl>plus"),
                (_("Zoom Out"), "<Ctrl>minus"),
                (_("Reset Zoom"), "<Ctrl>0"),
                (_("Pan Left"), "<Ctrl>Left"),
                (_("Pan Right"), "<Ctrl>Right"),
                (_("Pan Up"), "<Ctrl>Up"),
                (_("Pan Down"), "<Ctrl>Down"),
            ]),
            (_("General"), [
                (_("Toggle Sidebar"), "F9"),
                (_("Keyboard Shortcuts"), "<Ctrl>question"),
                (_("Preferences"), "<Ctrl>comma"),
                (_("Open Source Snippets"), "<Ctrl>P"),
		        (_("Delete Taken Screenshots"), "<Ctrl><Shift>D"),
            ])
        ]

        if self._has_adw_shortcuts_dialog():
            self.dialog = self._create_adw_dialog(sections_data)
        else:
            self.dialog = self._create_gtk_window(sections_data)

    def _create_adw_dialog(self, sections_data):
        dialog = Adw.ShortcutsDialog()

        for section_title, shortcuts in sections_data:
            section = Adw.ShortcutsSection(title=section_title)
            for title, accel in shortcuts:
                section.add(Adw.ShortcutsItem(title=title, accelerator=accel))
            dialog.add(section)

        return dialog

    def _create_gtk_window(self, sections_data):
        window = Gtk.ShortcutsWindow()

        main_section = Gtk.ShortcutsSection()
        main_section.set_visible(True)

        for section_title, shortcuts in sections_data:
            group = Gtk.ShortcutsGroup(title=section_title)

            for title, accel in shortcuts:
                shortcut = Gtk.ShortcutsShortcut(title=title, accelerator=accel)
                group.add_shortcut(shortcut)

            main_section.add_group(group)

        window.add_section(main_section)

        if self.parent:
            window.set_transient_for(self.parent)
            window.set_modal(True)

        return window

    def show(self):
        if not self.dialog:
            self.create()

        if self._has_adw_shortcuts_dialog() and hasattr(self.dialog, 'present'):
            self.dialog.present(self.parent)
        else:
            self.dialog.show()

    def set_parent(self, parent: Gtk.Window):
        self.parent = parent
        if self.dialog and hasattr(self.dialog, 'set_transient_for'):
            self.dialog.set_transient_for(parent)
            if hasattr(self.dialog, 'set_modal'):
                self.dialog.set_modal(True)



# Copyright (C) 2025 Alexander Vanhee, tfuxu
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import cairo
from gi.repository import Adw, Gdk, Gio, Gtk, GObject
from typing import Tuple
from enum import Enum

from gradia.overlay.drawing_actions import *
from gradia.overlay.text_entry_popover import TextEntryPopover

HANDLE_SIZE = 8

class ResizeHandle(Enum):
    NONE = "none"
    TOP_LEFT = "top_left"
    TOP_RIGHT = "top_right"
    BOTTOM_LEFT = "bottom_left"
    BOTTOM_RIGHT = "bottom_right"
    TOP = "top"
    BOTTOM = "bottom"
    LEFT = "left"
    RIGHT = "right"

    @classmethod
    def get_cursor_for_handle(cls, handle) -> str:
        cursor_map = {
            cls.TOP_LEFT: "nw-resize",
            cls.TOP_RIGHT: "ne-resize",
            cls.BOTTOM_LEFT: "sw-resize",
            cls.BOTTOM_RIGHT: "se-resize",
            cls.TOP: "ns-resize",
            cls.BOTTOM: "ns-resize",
            cls.LEFT: "ew-resize",
            cls.RIGHT: "ew-resize",
            cls.NONE: "grab"
        }
        return cursor_map.get(handle, "default")

class DrawingOverlay(Gtk.DrawingArea):
    __gtype_name__ = "GradiaDrawingOverlay"

    __gsignals__ = {
        'selection-changed': (GObject.SignalFlags.RUN_FIRST, None, (object,))
    }

    def __init__(self, **kwargs):
        super().__init__(can_focus=True, **kwargs)

        self.set_draw_func(self._on_draw)

        self.coordinate_transform = None
        self.delta_transform = None

        self.picture_widget = None
        self.options = None
        self.font_size = 22
        self.is_drawing = False
        self.current_stroke = []
        self.start_point = None
        self.end_point = None
        self.actions: list[DrawingAction] = []
        self.redo_stack = []
        self._next_number = 1

        self._selected_action: DrawingAction | None = None
        self.selection_start_pos = None
        self.is_moving_selection = False
        self.move_start_point = None
        self.current_shift_pressed = False

        self.is_resizing = False
        self.resize_handle = ResizeHandle.NONE
        self.resize_start_bounds = None
        self.resize_start_mouse = None

        self.text_entry_popup = None
        self.text_position = None
        self.is_text_editing = False
        self.live_text = None
        self.editing_text_action = None

        self._setup_gestures()

    def set_picture_reference(self, picture: Gtk.Picture) -> None:
        self.picture_widget = picture
        picture.connect("notify::paintable", lambda *args: self.queue_draw())

    def set_erase_selected_revealer(self, erase_selected_revealer: Gtk.Revealer) -> None:
        self.erase_selected_revealer = erase_selected_revealer

    @property
    def selected_action(self) -> DrawingAction | None:
        return self._selected_action

    @selected_action.setter
    def selected_action(self, action: DrawingAction | None) -> None:
        self._selected_action = action
        self.erase_selected_revealer.set_reveal_child(action is not None)
        self.emit('selection-changed', action.options if action else None)

    def _can_resize_action(self, action: DrawingAction) -> bool:
        return isinstance(action, (RectAction, CircleAction, CensorAction, ArrowAction, LineAction))

    def _get_resize_handles(self, action: DrawingAction) -> list:
        if not self._can_resize_action(action):
            return []

        if isinstance(action, (ArrowAction, LineAction)):
            start_x_widget, start_y_widget = self._image_to_widget_coords(*action.start)
            end_x_widget, end_y_widget = self._image_to_widget_coords(*action.end)

            handles = [
                (ResizeHandle.TOP_LEFT, start_x_widget - HANDLE_SIZE/2, start_y_widget - HANDLE_SIZE/2),
                (ResizeHandle.BOTTOM_RIGHT, end_x_widget - HANDLE_SIZE/2, end_y_widget - HANDLE_SIZE/2),
            ]
            return handles
        else:
            min_x_img, min_y_img, max_x_img, max_y_img = action.get_bounds().get_bounding_rect()
            x1_widget, y1_widget = self._image_to_widget_coords(min_x_img, min_y_img)
            x2_widget, y2_widget = self._image_to_widget_coords(max_x_img, max_y_img)

            handles = [
                (ResizeHandle.TOP_LEFT, x1_widget - HANDLE_SIZE/2, y1_widget - HANDLE_SIZE/2),
                (ResizeHandle.TOP_RIGHT, x2_widget - HANDLE_SIZE/2, y1_widget - HANDLE_SIZE/2),
                (ResizeHandle.BOTTOM_LEFT, x1_widget - HANDLE_SIZE/2, y2_widget - HANDLE_SIZE/2),
                (ResizeHandle.BOTTOM_RIGHT, x2_widget - HANDLE_SIZE/2, y2_widget - HANDLE_SIZE/2),
            ]
            return handles

    def _get_handle_at_point(self, x_widget: float, y_widget: float) -> ResizeHandle:
        if not self.selected_action or not self._can_resize_action(self.selected_action):
            return ResizeHandle.NONE

        handles = self._get_resize_handles(self.selected_action)
        for handle_type, handle_x, handle_y in handles:
            if (handle_x <= x_widget <= handle_x + HANDLE_SIZE and
                handle_y <= y_widget <= handle_y + HANDLE_SIZE):
                return handle_type

        if isinstance(self.selected_action, (ArrowAction, LineAction)):
            return ResizeHandle.NONE

        min_x_img, min_y_img, max_x_img, max_y_img = self.selected_action.get_bounds().get_bounding_rect()
        x1_widget, y1_widget = self._image_to_widget_coords(min_x_img, min_y_img)
        x2_widget, y2_widget = self._image_to_widget_coords(max_x_img, max_y_img)

        margin = HANDLE_SIZE

        if abs(y_widget - y1_widget) <= margin and x1_widget <= x_widget <= x2_widget:
            return ResizeHandle.TOP
        if abs(y_widget - y2_widget) <= margin and x1_widget <= x_widget <= x2_widget:
            return ResizeHandle.BOTTOM
        if abs(x_widget - x1_widget) <= margin and y1_widget <= y_widget <= y2_widget:
            return ResizeHandle.LEFT
        if abs(x_widget - x2_widget) <= margin and y1_widget <= y_widget <= y2_widget:
            return ResizeHandle.RIGHT

        return ResizeHandle.NONE

    def _resize_action(
        self,
        action: DrawingAction,
        handle: ResizeHandle,
        start_bounds: tuple,
        start_mouse: tuple,
        current_mouse: tuple,
        shift_pressed: bool
    ):
        if isinstance(action, (ArrowAction, LineAction)):
            current_mouse_x, current_mouse_y = current_mouse

            if handle == ResizeHandle.TOP_LEFT:
                action.start = (int(current_mouse_x), int(current_mouse_y))
            elif handle == ResizeHandle.BOTTOM_RIGHT:
                action.end = (int(current_mouse_x), int(current_mouse_y))
        else:
            min_x_start, min_y_start, max_x_start, max_y_start = start_bounds
            start_mouse_x, start_mouse_y = start_mouse
            current_mouse_x, current_mouse_y = current_mouse

            delta_x = current_mouse_x - start_mouse_x
            delta_y = current_mouse_y - start_mouse_y

            if handle in [ResizeHandle.TOP_LEFT, ResizeHandle.TOP, ResizeHandle.TOP_RIGHT]:
                min_y_start += delta_y
            if handle in [ResizeHandle.BOTTOM_LEFT, ResizeHandle.BOTTOM, ResizeHandle.BOTTOM_RIGHT]:
                max_y_start += delta_y
            if handle in [ResizeHandle.TOP_LEFT, ResizeHandle.LEFT, ResizeHandle.BOTTOM_LEFT]:
                min_x_start += delta_x
            if handle in [ResizeHandle.TOP_RIGHT, ResizeHandle.RIGHT, ResizeHandle.BOTTOM_RIGHT]:
                max_x_start += delta_x

            action.start = (int(min_x_start), int(min_y_start))
            action.end = (int(max_x_start), int(max_y_start))

    def _get_image_bounds(self) -> Tuple[float, float, float, float]:
        if not self.picture_widget or not self.picture_widget.get_paintable():
            return 0, 0, float(self.get_width()), float(self.get_height())

        widget_w = float(self.picture_widget.get_width())
        widget_h = float(self.picture_widget.get_height())
        img_w_intrinsic = float(self.picture_widget.get_paintable().get_intrinsic_width())
        img_h_intrinsic = float(self.picture_widget.get_paintable().get_intrinsic_height())

        if img_w_intrinsic <= 0 or img_h_intrinsic <= 0:
            return 0, 0, widget_w, widget_h

        scale = min(widget_w / img_w_intrinsic, widget_h / img_h_intrinsic)

        disp_w = img_w_intrinsic * scale
        disp_h = img_h_intrinsic * scale

        offset_x = (widget_w - disp_w) / 2
        offset_y = (widget_h - disp_h) / 2
        return offset_x, offset_y, disp_w, disp_h

    def _get_modified_image_bounds(self) -> Tuple[int, int]:
        if not self.picture_widget or not self.picture_widget.get_paintable():
            return 0, 0
        return self.picture_widget.get_paintable().get_intrinsic_width(), \
               self.picture_widget.get_paintable().get_intrinsic_height()

    def _get_scale_factor(self) -> float:
        _, _, dw, dh = self._get_image_bounds()
        if not self.picture_widget or not self.picture_widget.get_paintable():
            return 1.0
        img_w_intrinsic = self.picture_widget.get_paintable().get_intrinsic_width()
        return dw / img_w_intrinsic if img_w_intrinsic else 1.0

    def _widget_to_image_coords(self, x_widget: float, y_widget: float) -> Tuple[int, int]:
        ox, oy, disp_w, disp_h = self._get_image_bounds()
        scale = self._get_scale_factor()

        rel_x_on_disp_image = x_widget - ox
        rel_y_on_disp_image = y_widget - oy

        img_x_intrinsic_top_left = rel_x_on_disp_image / scale
        img_y_intrinsic_top_left = rel_y_on_disp_image / scale

        img_w_intrinsic, img_h_intrinsic = self._get_modified_image_bounds()

        center_x_intrinsic = img_w_intrinsic / 2
        center_y_intrinsic = img_h_intrinsic / 2

        img_x_centered = round(img_x_intrinsic_top_left - center_x_intrinsic)
        img_y_centered = round(img_y_intrinsic_top_left - center_y_intrinsic)

        return img_x_centered, img_y_centered

    def _image_to_widget_coords(self, x_image: int, y_image: int) -> Tuple[float, float]:
        ox, oy, disp_w, disp_h = self._get_image_bounds()
        scale = self._get_scale_factor()

        img_w_intrinsic, img_h_intrinsic = self._get_modified_image_bounds()

        center_x_intrinsic = img_w_intrinsic / 2
        center_y_intrinsic = img_h_intrinsic / 2

        img_x_intrinsic_top_left = center_x_intrinsic + x_image
        img_y_intrinsic_top_left = center_y_intrinsic + y_image

        rel_x_on_disp_image = img_x_intrinsic_top_left * scale
        rel_y_on_disp_image = img_y_intrinsic_top_left * scale

        widget_x = ox + rel_x_on_disp_image
        widget_y = oy + rel_y_on_disp_image

        return widget_x, widget_y

    def _is_point_in_image(self, x_widget: float, y_widget: float) -> bool:
        ox, oy, dw, dh = self._get_image_bounds()
        return ox <= x_widget <= ox + dw and oy <= y_widget <= oy + dh

    def _get_background_pixbuf(self):
        if not self.picture_widget:
            return None

        paintable = self.picture_widget.get_paintable()
        if isinstance(paintable, Gdk.Texture):
            return Gdk.pixbuf_get_from_texture(paintable)
        return None

    def _setup_actions(self):
        for mode in DrawingMode:
            action = Gio.SimpleAction.new(f"drawing-mode-{mode.value}", None)
            action.connect("activate", lambda a, p, m=mode: self.set_drawing_mode(m))
            root = self.get_root()
            if hasattr(root, "add_action"):
                root.add_action(action)

    def _get_number_actions(self) -> list:
        return [action for action in self.actions if isinstance(action, NumberStampAction)]

    def _renumber_actions(self):
        number_actions = self._get_number_actions()
        number_actions.sort(key=lambda action: action.creation_time)

        for i, action in enumerate(number_actions, 1):
            action.number = i
        self._next_number = len(number_actions) + 1

    def remove_selected_action(self) -> bool:
        if self.selected_action and self.selected_action in self.actions:
            was_number_action = isinstance(self.selected_action, NumberStampAction)
            self.actions.remove(self.selected_action)
            self.selected_action = None
            self.redo_stack.clear()

            if was_number_action:
                self._renumber_actions()

            self._update_undo_redo_action_states()
            self.queue_draw()
            return True
        return False

    def set_drawing_mode(self, mode: DrawingMode) -> None:
        if self.text_entry_popup:
            self._close_text_entry()

        if mode != DrawingMode.SELECT:
            self.selected_action = None

        self.options.mode = mode
        self.is_drawing = False
        self.is_moving_selection = False
        self.is_resizing = False
        self.resize_handle = ResizeHandle.NONE
        self.current_stroke.clear()
        self.start_point = None
        self.end_point = None
        self.queue_draw()

    def _find_action_at_point(self, x_image: int, y_image: int) -> DrawingAction | None:
        for action in reversed(self.actions):
            if action.contains_point(x_image, y_image):
                return action
        return None


    def _is_point_in_selection_bounds(self, x_image: int, y_image: int) -> bool:
        if not self.selected_action:
            return False

        min_x, min_y, max_x, max_y = self.selected_action.get_bounds().get_bounding_rect()
        padding_img = max(self.options.size, self.font_size / 2)

        return min_x - padding_img <= x_image <= max_x + padding_img and \
               min_y - padding_img <= y_image <= max_y + padding_img

    def _draw_selection_box(self, cr: cairo.Context, scale: float):
        if not self.selected_action:
            return

        bounds = self.selected_action.get_bounds()
        points = bounds.get_points()
        widget_points = [self._image_to_widget_coords(int(p[0]), int(p[1])) for p in points]
        accent = Adw.StyleManager.get_default().get_accent_color_rgba()
        cr.set_source_rgba(*accent)
        cr.set_line_width(2)
        cr.move_to(*widget_points[0])
        for point in widget_points[1:]:
            cr.line_to(*point)
        cr.close_path()
        cr.stroke()
        if self._can_resize_action(self.selected_action):
            handles = self._get_resize_handles(self.selected_action)
            for handle_type, handle_x, handle_y in handles:
                cr.set_source_rgba(1.0, 1.0, 1.0, 1.0)
                cr.rectangle(handle_x, handle_y, HANDLE_SIZE, HANDLE_SIZE)
                cr.fill()
                cr.set_source_rgba(*accent)
                cr.rectangle(handle_x, handle_y, HANDLE_SIZE, HANDLE_SIZE)
                cr.stroke()

    def _setup_gestures(self):
        click = Gtk.GestureClick.new()
        click.set_button(1)
        click.connect("pressed", self._on_click)
        self.add_controller(click)

        drag = Gtk.GestureDrag.new()
        drag.set_button(1)
        drag.connect("drag-begin", self._on_drag_begin)
        drag.connect("drag-update", self._on_drag_update)
        drag.connect("drag-end", self._on_drag_end)
        self.add_controller(drag)

        motion = Gtk.EventControllerMotion.new()
        motion.connect("motion", self._on_motion)
        self.add_controller(motion)

    def update_shift_state(self, gesture):
        state = gesture.get_current_event_state()
        shift_pressed = bool(state & Gdk.ModifierType.SHIFT_MASK)
        self.current_shift_pressed = shift_pressed

    def _on_click(self, gesture, n_press, x_widget, y_widget):
        original_x, original_y = x_widget, y_widget
        x_widget, y_widget = self.coordinate_transform(x_widget, y_widget)
        if self.options.mode == DrawingMode.TEXT and self._is_point_in_image(x_widget, y_widget):
            self.grab_focus()
            if n_press == 1:
                self._show_text_entry(original_x,original_y)
        elif self.options.mode == DrawingMode.NUMBER and self._is_point_in_image(x_widget, y_widget) and n_press == 1:
            self.grab_focus()
            img_x, img_y = self._widget_to_image_coords(x_widget, y_widget)
            number_action = NumberStampAction(
                position=(img_x, img_y),
                number=self._next_number,
                options=self.options.copy()
            )

            self.actions.append(number_action)
            self._renumber_actions()
            self.redo_stack.clear()
            self._update_undo_redo_action_states()
            self.queue_draw()

        elif self.options.mode == DrawingMode.SELECT and self._is_point_in_image(x_widget, y_widget):
            self.grab_focus()
            img_x, img_y = self._widget_to_image_coords(x_widget, y_widget)

            if (n_press == 2 and
                self.selected_action and
                isinstance(self.selected_action, TextAction) and
                self.selected_action.contains_point(img_x, img_y)):
                self._start_text_edit(self.selected_action,original_x,original_y)
                return

            if n_press == 1:
                handle = self._get_handle_at_point(x_widget, y_widget)
                if handle != ResizeHandle.NONE and self.selected_action and self._can_resize_action(self.selected_action):
                    return

                if self.selected_action and not self._is_point_in_selection_bounds(img_x, img_y):
                    self.selected_action = None
                    self.queue_draw()

                action = self._find_action_at_point(img_x, img_y)
                if action and action != self.selected_action:
                    self.selected_action = action
                    self.queue_draw()
                elif not action and self.selected_action:
                    self.selected_action = None
                    self.queue_draw()

    def _start_text_edit(self, text_action, widget_x, widget_y):
        self.editing_text_action = text_action
        self.text_position = text_action.position
        self.is_text_editing = True
        self.live_text = text_action.text
        self.text_entry_popup = TextEntryPopover(
            parent=self,
            on_text_activate=self._on_text_entry_activate,
            on_text_changed=self._on_text_entry_changed,
            on_font_size_changed=self._on_font_size_changed,
            font_size=text_action.font_size,
            initial_text=text_action.text
        )
        self.text_entry_popup.connect("closed", self._on_text_entry_popover_closed)
        self.text_entry_popup.popup_at_widget_coords(self, widget_x, widget_y)

    def _show_text_entry(self, x_widget, y_widget):
        original_x, original_y = x_widget, y_widget
        x_widget, y_widget = self.coordinate_transform(x_widget, y_widget)

        if self.text_entry_popup:
            self.text_entry_popup.popdown()
            self.text_entry_popup = None

        self.text_position = self._widget_to_image_coords(x_widget, y_widget)
        self.is_text_editing = True
        self.live_text = ""
        self.editing_text_action = None
        self.text_entry_popup = TextEntryPopover(
            parent=self,
            on_text_activate=self._on_text_entry_activate,
            on_text_changed=self._on_text_entry_changed,
            on_font_size_changed=self._on_font_size_changed,
            font_size=self.font_size
        )
        self.text_entry_popup.connect("closed", self._on_text_entry_popover_closed)
        self.text_entry_popup.popup_at_widget_coords(self, original_x, original_y)

    def _on_font_size_changed(self, spin_button):
        font_size = spin_button.get_value()
        if self.editing_text_action:
            self.editing_text_action.font_size = font_size
            self.editing_text_action.font_size = font_size
            if self.selected_action == self.editing_text_action:
                self.queue_draw()
        else:
            self.font_size = font_size

        if self.live_text:
            self.queue_draw()

    def _on_text_entry_popover_closed(self, popover):
        if self.text_entry_popup and self.text_position:
            text = self.text_entry_popup.get_text().strip()

            if self.editing_text_action:
                if text:
                    self.editing_text_action.text = text
                    if hasattr(self.text_entry_popup, 'spin'):
                        self.editing_text_action.font_size = self.text_entry_popup.spin.get_value()
                else:
                    if self.editing_text_action in self.actions:
                        self.actions.remove(self.editing_text_action)
                    if self.selected_action == self.editing_text_action:
                        self.selected_action = None
                self.redo_stack.clear()
                self._update_undo_redo_action_states()
            else:
                if text:
                    current_settings = self.options.copy()
                    if hasattr(self.text_entry_popup, 'spin'):
                        current_settings.font_size = self.text_entry_popup.spin.get_value()

                    action = TextAction(
                        self.text_position,
                        text,
                        self._get_modified_image_bounds(),
                        current_settings,
                        self.font_size
                    )
                    self.actions.append(action)
                    self.redo_stack.clear()
                    self._update_undo_redo_action_states()

        self._cleanup_text_entry()
        self.queue_draw()

    def _on_text_entry_changed(self, entry):
        if not self.text_entry_popup:
            return
        self.live_text = self.text_entry_popup.get_text().strip()
        if self.editing_text_action:
            self.editing_text_action.text = self.live_text

        self.queue_draw()

    def _on_text_entry_activate(self, entry):
        self._close_text_entry()
        self.queue_draw()

    def _cleanup_text_entry(self):
        if self.text_entry_popup:
            self.text_entry_popup = None
        self.text_position = None
        self.live_text = None
        self.is_text_editing = False
        self.editing_text_action = None

    def _close_text_entry(self):
        if self.text_entry_popup:
            self.text_entry_popup.popdown()
            self.text_entry_popup = None
        self.text_position = None
        self.live_text = None
        self.is_text_editing = False
        self.editing_text_action = None

    def _on_drag_begin(self, gesture, x_widget, y_widget):
        x_widget, y_widget = self.coordinate_transform(x_widget, y_widget)
        if self.options.mode == DrawingMode.TEXT or self.options.mode == DrawingMode.NUMBER or self.text_entry_popup:
            return
        if not self._is_point_in_image(x_widget, y_widget):
            return

        self.grab_focus()
        img_x, img_y = self._widget_to_image_coords(x_widget, y_widget)

        self.update_shift_state(gesture)

        if self.options.mode == DrawingMode.SELECT:
            if self.selected_action:
                handle = self._get_handle_at_point(x_widget, y_widget)
                if handle != ResizeHandle.NONE and self._can_resize_action(self.selected_action):
                    self.is_resizing = True
                    self.resize_handle = handle
                    self.resize_start_bounds = self.selected_action.get_bounds().get_bounding_rect()
                    self.resize_start_mouse = (img_x, img_y)
                    return
                elif self._is_point_in_selection_bounds(img_x, img_y):
                    self.is_moving_selection = True
                    self.move_start_point = (img_x, img_y)
                    return

            self.selected_action = self._find_action_at_point(img_x, img_y)
            if self.selected_action:
                self.is_moving_selection = True
                self.move_start_point = (img_x, img_y)
            self.queue_draw()
            return

        self.is_drawing = True
        if self.options.mode == DrawingMode.PEN or self.options.mode == DrawingMode.HIGHLIGHTER:
            self.current_stroke = [(img_x, img_y)]
        else:
            self.start_point = (img_x, img_y)
            self.end_point = (img_x, img_y)

    def _on_drag_update(self, gesture, dx_widget, dy_widget):
        dx_widget, dy_widget = self.delta_transform(dx_widget, dy_widget)
        if self.options.mode == DrawingMode.TEXT or self.options.mode == DrawingMode.NUMBER:
            return

        start_x_raw, start_y_raw = gesture.get_start_point().x, gesture.get_start_point().y
        start_x_widget, start_y_widget = self.coordinate_transform(start_x_raw, start_y_raw)
        cur_x_widget, cur_y_widget = start_x_widget + dx_widget, start_y_widget + dy_widget
        img_x, img_y = self._widget_to_image_coords(cur_x_widget, cur_y_widget)

        self.update_shift_state(gesture)

        if self.options.mode == DrawingMode.SELECT and self.is_resizing and self.selected_action and self.resize_start_bounds:
            self._resize_action(self.selected_action, self.resize_handle, self.resize_start_bounds,
                              self.resize_start_mouse, (img_x, img_y), self.current_shift_pressed)
            self.queue_draw()
            return

        if self.options.mode == DrawingMode.SELECT and self.is_moving_selection and self.selected_action and self.move_start_point:
            old_x_img, old_y_img = self.move_start_point
            delta_x_img = img_x - old_x_img
            delta_y_img = img_y - old_y_img
            self.selected_action.translate(delta_x_img, delta_y_img)
            self.move_start_point = (img_x, img_y)
            self.queue_draw()
            return

        if not self.is_drawing:
            return

        if self.options.mode == DrawingMode.PEN or self.options.mode == DrawingMode.HIGHLIGHTER:
            self.current_stroke.append((img_x, img_y))
        else:
            self.end_point = (img_x, img_y)
        self.queue_draw()

    def _on_drag_end(self, gesture, dx_widget, dy_widget):
        dx_widget, dy_widget = self.delta_transform(dx_widget, dy_widget)
        if self.options.mode == DrawingMode.TEXT or self.options.mode == DrawingMode.NUMBER:
            return

        if self.options.mode == DrawingMode.SELECT:
            if self.is_resizing:
                self.is_resizing = False
                self.resize_handle = ResizeHandle.NONE
                self.resize_start_bounds = None
                self.resize_start_mouse = None
                self.redo_stack.clear()
                self._update_undo_redo_action_states()
                return
            self.is_moving_selection = False
            self.move_start_point = None
            return

        if not self.is_drawing:
            return

        self.update_shift_state(gesture)

        self.is_drawing = False
        mode = self.options.mode
        if (mode == DrawingMode.PEN or mode == DrawingMode.HIGHLIGHTER) and len(self.current_stroke) > 1:
            if mode == DrawingMode.PEN:
                self.actions.append(StrokeAction(self.current_stroke.copy(), self.options.copy()))
            else:
                self.actions.append(HighlighterAction(self.current_stroke.copy(), self.options.copy(), self.current_shift_pressed))
            self.current_stroke.clear()
        elif self.start_point and self.end_point:
            if mode == DrawingMode.ARROW:
                self.actions.append(ArrowAction(self.start_point, self.end_point,self.current_shift_pressed, self.options.copy()))
            elif mode == DrawingMode.LINE:
                self.actions.append(LineAction(self.start_point, self.end_point,self.current_shift_pressed, self.options.copy()))
            elif mode == DrawingMode.SQUARE:
                self.actions.append(RectAction(self.start_point, self.end_point,self.current_shift_pressed, self.options.copy()))
            elif mode == DrawingMode.CIRCLE:
                self.actions.append(CircleAction(self.start_point, self.end_point,self.current_shift_pressed, self.options.copy()))
            elif mode == DrawingMode.CENSOR:
                censor_action = CensorAction(self.start_point, self.end_point, self._get_background_pixbuf(), self.options.copy())
                current_scale = self._get_scale_factor()
                censor_action.set_original_scale(current_scale)
                self.actions.append(censor_action)

        self.start_point = None
        self.end_point = None
        self.redo_stack.clear()
        self._update_undo_redo_action_states()
        self.queue_draw()

    def _on_motion(self, controller, x_widget, y_widget):
        x_widget, y_widget = self.coordinate_transform(x_widget, y_widget)

        if self.options.mode == DrawingMode.TEXT:
            name = "text" if self._is_point_in_image(x_widget, y_widget) else "default"

        elif self.options.mode == DrawingMode.NUMBER:
            name = "crosshair" if self._is_point_in_image(x_widget, y_widget) else "default"

        elif self.options.mode == DrawingMode.CENSOR:
            name = "crosshair" if self._is_point_in_image(x_widget, y_widget) else "default"

        elif self.options.mode == DrawingMode.SELECT:
            name = self._get_select_mode_cursor(x_widget, y_widget)

        else:
            if self._is_point_in_image(x_widget, y_widget):
                name = "crosshair"
            else:
                name = "default"

        self.set_cursor(Gdk.Cursor.new_from_name(name, None))

    def _get_select_mode_cursor(self, x_widget, y_widget):
        if self.is_resizing or self.is_moving_selection:
            if self.is_resizing:
                if isinstance(self.selected_action, (ArrowAction, LineAction)):
                    return "move"
                else:
                    return ResizeHandle.get_cursor_for_handle(self.resize_handle)
            else:
                return "grab"

        img_x, img_y = self._widget_to_image_coords(x_widget, y_widget)

        if self.selected_action:
            handle = self._get_handle_at_point(x_widget, y_widget)
            if handle != ResizeHandle.NONE and self._can_resize_action(self.selected_action):
                if isinstance(self.selected_action, (ArrowAction, LineAction)):
                    return "move"
                else:
                    return ResizeHandle.get_cursor_for_handle(handle)
            elif self._is_point_in_selection_bounds(img_x, img_y):
                return "grab"
            else:
                return "pointer" if self._find_action_at_point(img_x, img_y) else "default"
        else:
            return "pointer" if self._find_action_at_point(img_x, img_y) else "default"

    def _on_draw(self, area, cr: cairo.Context, width: int, height: int):
        scale = self._get_scale_factor()
        ox, oy, dw, dh = self._get_image_bounds()

        cr.set_line_cap(cairo.LineCap.ROUND)
        cr.set_line_join(cairo.LineJoin.ROUND)
        cr.rectangle(ox, oy, dw, dh)
        cr.clip()

        for action in self.actions:
            if action == self.editing_text_action and self.is_text_editing:
                continue
            action.draw(cr, self._image_to_widget_coords, scale)

        if self.is_drawing and self.options.mode != DrawingMode.TEXT and self.options.mode != DrawingMode.NUMBER:
            cr.set_source_rgba(*self.options.primary_color)
            if self.options.mode == DrawingMode.PEN and len(self.current_stroke) > 1:
                StrokeAction(self.current_stroke, self.options.copy()).draw(cr, self._image_to_widget_coords, scale)
            elif self.options.mode == DrawingMode.HIGHLIGHTER and len(self.current_stroke) > 1:
                HighlighterAction(self.current_stroke, self.options.copy(), self.current_shift_pressed).draw(cr, self._image_to_widget_coords, scale)
            elif self.start_point and self.end_point:
                if self.options.mode == DrawingMode.ARROW:
                    ArrowAction(self.start_point, self.end_point,self.current_shift_pressed, self.options.copy()).draw(cr, self._image_to_widget_coords, scale)
                elif self.options.mode == DrawingMode.LINE:
                    LineAction(self.start_point, self.end_point,self.current_shift_pressed, self.options.copy()).draw(cr, self._image_to_widget_coords, scale)
                elif self.options.mode == DrawingMode.SQUARE:
                    RectAction(self.start_point, self.end_point, self.current_shift_pressed, self.options.copy()).draw(cr, self._image_to_widget_coords, scale)
                elif self.options.mode == DrawingMode.CIRCLE:
                    CircleAction(self.start_point, self.end_point, self.current_shift_pressed, self.options.copy()).draw(cr, self._image_to_widget_coords, scale)
                elif self.options.mode == DrawingMode.CENSOR:
                    temp_censor = CensorAction(self.start_point, self.end_point, self._get_background_pixbuf(), self.options.copy())
                    temp_censor.set_original_scale(scale)
                    temp_censor.draw(cr, self._image_to_widget_coords, scale)

        if self.is_text_editing and self.text_position and self.live_text:
            if self.editing_text_action:
                preview = TextAction(
                    self.text_position,
                    self.live_text,
                    self._get_modified_image_bounds(),
                    self.editing_text_action.options.copy(),
                    self.editing_text_action.font_size
                )
            else:
                preview = TextAction(
                    self.text_position,
                    self.live_text,
                    self._get_modified_image_bounds(),
                    self.options.copy(),
                    self.font_size
                )
            preview.draw(cr, self._image_to_widget_coords, scale)

        if self.selected_action:
            self._draw_selection_box(cr, scale)

    def export_to_pixbuf(self, requested_width, requested_height) -> GdkPixbuf.Pixbuf | None:
        if not self.picture_widget or not self.picture_widget.get_paintable():
            return None

        paintable = self.picture_widget.get_paintable()
        img_w = paintable.get_intrinsic_width()
        img_h = paintable.get_intrinsic_height()

        scale_factor_x = requested_width / img_w
        scale_factor_y = requested_height / img_h

        return render_actions_to_pixbuf(self.actions, requested_width, requested_height, scale_factor_x, scale_factor_y)

    def clear_drawing(self) -> None:
        self._close_text_entry()
        self.actions.clear()
        self.redo_stack.clear()
        self.selected_action = None
        self._next_number = 1
        self._update_undo_redo_action_states()
        self.queue_draw()

    def undo(self) -> None:
        if self.actions:
            undone_action = self.actions.pop()
            self.redo_stack.append(undone_action)
            self.selected_action = None

            if isinstance(undone_action, NumberStampAction):
                self._renumber_actions()

            self._update_undo_redo_action_states()
            self.queue_draw()

    def redo(self) -> None:
        if self.redo_stack:
            redone_action = self.redo_stack.pop()
            self.actions.append(redone_action)
            self.selected_action = None

            if isinstance(redone_action, NumberStampAction):
                self._renumber_actions()

            self._update_undo_redo_action_states()
            self.queue_draw()

    def _update_undo_redo_action_states(self) -> None:
        root = self.get_root()
        if root:
            undo_action = root.lookup_action("undo")
            if undo_action:
                undo_action.set_enabled(bool(self.actions))
            
            redo_action = root.lookup_action("redo")
            if redo_action:
                redo_action.set_enabled(bool(self.redo_stack))


    def set_drawing_visible(self, is_visible: bool) -> None:
        self.set_visible(is_visible)

    def get_drawing_visible(self) -> bool:
        return self.get_visible()


def render_actions_to_pixbuf(actions: list[DrawingAction], width: int, height: int, scale_factor_x: float = 1.0, scale_factor_y: float = 1.0) -> GdkPixbuf.Pixbuf | None:
    if width <= 0 or height <= 0:
        return None

    surface = cairo.ImageSurface(cairo.Format.ARGB32, width, height)
    cr = cairo.Context(surface)

    cr.set_operator(cairo.Operator.CLEAR)
    cr.paint()
    cr.set_operator(cairo.Operator.OVER)

    def image_coords_to_intrinsic_pixels(x_image: int, y_image: int) -> Tuple[float, float]:
        center_x_intrinsic = width / 2.0
        center_y_intrinsic = height / 2.0
        return (center_x_intrinsic + x_image * scale_factor_x, center_y_intrinsic + y_image * scale_factor_y)

    cr.set_line_cap(cairo.LineCap.ROUND)
    cr.set_line_join(cairo.LineJoin.ROUND)

    scale_factor = (scale_factor_x + scale_factor_y) / 2.0

    for action in actions:
        action.draw(cr, image_coords_to_intrinsic_pixels, scale_factor)

    surface.flush()

    return Gdk.pixbuf_get_from_surface(surface, 0, 0, width, height)



# Copyright (C) 2025 Alexander Vanhee
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Adw, Gio, GLib, Gtk
from typing import Callable, Optional
from gradia.backend.settings import Settings
from gradia.constants import rootdir  # pyright: ignore

@Gtk.Template(resource_path=f"{rootdir}/ui/confirm_close_dialog.ui")
class ConfirmCloseDialog(Adw.AlertDialog):
    __gtype_name__ = "GradiaConfirmCloseDialog"

    dont_ask_switch: Gtk.CheckButton = Gtk.Template.Child()

    def __init__(self, parent_window: Adw.ApplicationWindow, **kwargs) -> None:
        super().__init__(**kwargs)
        self.parent_window = parent_window
        self.on_confirm_callback: Optional[Callable[[], None]] = None
        self.on_copy_callback: Optional[Callable[[], None]] = None
        self.settings = Settings()
        self._setup_settings_binding()

    """
    Setup Methods
    """

    def _setup_settings_binding(self) -> None:
        self.dont_ask_switch.set_active(self.settings.exit_method == "none")

        def on_switch_toggled(switch: Gtk.CheckButton):
            if switch.get_active():
                self.settings.exit_method = "none"
            else:
                self.settings.exit_method = "confirm"

        self.dont_ask_switch.connect("toggled", on_switch_toggled)

    """
    Public Methods
    """

    def show_dialog(self, on_confirm_callback: Callable[[], None], on_copy_callback: Callable[[], None]) -> None:
        """Show the confirmation dialog"""
        self.on_confirm_callback = on_confirm_callback
        self.on_copy_callback = on_copy_callback
        self.choose(self.parent_window, None, self._on_response)

    """
    Callbacks
    """

    def _on_response(self, dialog: Adw.AlertDialog, task: Gio.Task) -> None:
        response = dialog.choose_finish(task)
        if response == "close" and self.on_confirm_callback:
            self.on_confirm_callback()
        if response == "copy" and self.on_copy_callback:
            self.on_copy_callback()



# Copyright (C) 2025 Alexander Vanhee
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import sys
import os
import tempfile
import shutil

from collections.abc import Sequence
from typing import Optional

from PIL import Image

from gi.repository import Adw, Gio, Xdp

from gradia.constants import app_id, rootdir  # pyright: ignore
from gradia.ui.window import GradiaMainWindow
from gradia.ui.dialog.ocr_launcher import present_ocr_dialog
from gradia.backend.logger import Logger
from gradia.backend.ocr import OCR
from gradia.utils.std_image_loader import StdinImageLoader

logging = Logger()


class GradiaApp(Adw.Application):
    __gtype_name__ = "GradiaApp"

    def __init__(self, version: str):
        super().__init__(
            application_id=app_id,
            flags=Gio.ApplicationFlags.HANDLES_COMMAND_LINE | Gio.ApplicationFlags.HANDLES_OPEN
        )

        self.set_resource_base_path(rootdir)

        self.version = version
        self.temp_dirs: list[str] = []
        self._stdin_image_path: Optional[str] = None


        self.connect("shutdown", self.on_shutdown)

    def do_command_line(self, command_line: Gio.ApplicationCommandLine) -> int:
        args = command_line.get_arguments()[1:]
        logging.info(f"Command line arguments: {args}")

        if "--help" in args or "-h" in args:
            self._print_help()
            return 0

        files_to_open = []
        screenshot_file = None
        ocr_file = None

        for arg in args:
            if arg.startswith("--screenshot-file="):
                screenshot_file = arg.split("=", 1)[1]
                logging.info(f"Screenshot file detected: {screenshot_file}")
            elif arg.startswith("--ocr-file="):
                ocr_file = arg.split("=", 1)[1]
                logging.info(f"OCR file detected: {ocr_file}")
            elif not arg.startswith("--"):
                try:
                    file = Gio.File.new_for_commandline_arg(arg)
                    path = file.get_path()
                    if path:
                        files_to_open.append(path)
                        logging.debug(f"File to open detected: {path}")
                    else:
                        logging.warning(f"Argument {arg} does not have a valid path.")
                except Exception as e:
                    logging.warning(f"Failed to parse file URI {arg}.", exception=e, show_exception=True)

        if ocr_file:
            self._open_ocr_standalone(ocr_file)
        elif files_to_open:
            for path in files_to_open:
                self._open_window(file_path=path)
        elif screenshot_file:
            self._open_window(start_screenshot=screenshot_file)
        else:
            self.activate()

        return 0

    def _print_help(self):
        file = Gio.File.new_for_uri("resource:///be/alexandervanhee/gradia/help.txt")
        stream = file.read(None)
        contents = stream.read_bytes(4096, None).get_data().decode("utf-8")
        print(contents)

    def do_open(self, files: Sequence[Gio.File], hint: str):
        logging.debug(f"do_open called with files: {[file.get_path() for file in files]} and hint: {hint}")
        for file in files:
            path = file.get_path()
            if path:
                self._open_window(path)

    def do_activate(self):
        logging.debug("do_activate called")

        stdin_path = self._stdin_image_path
        self._stdin_image_path = None

        if stdin_path:
            logging.debug(f"Opening window with stdin image path: {stdin_path}")
            self._open_window(stdin_path)
        else:
            self._open_window(None)

    def _open_window(
        self,
        file_path: Optional[str] = None,
        start_screenshot: Optional[str] = None,
    ):
        logging.info(f"Opening window with file_path={file_path}")
        temp_dir = tempfile.mkdtemp()
        logging.debug(f"Created temp directory: {temp_dir}")
        self.temp_dirs.append(temp_dir)

        window = GradiaMainWindow(
            temp_dir=temp_dir,
            version=self.version,
            application=self,
            file_path=file_path,
            start_screenshot=start_screenshot,
        )
        window.show()

    def _open_ocr_standalone(self, image_path: str) -> None:
        logging.info(f"Opening standalone OCR dialog for: {image_path}")

        if not os.path.isfile(image_path):
            logging.warning(f"OCR file does not exist: {image_path}")
            return

        try:
            pil_image = Image.open(image_path)
            pil_image.load()
        except Exception as e:
            logging.warning(f"Failed to load image for OCR: {image_path}", exception=e, show_exception=True)
            return

        self.hold()
        released = {"done": False}

        def release_once():
            if not released["done"]:
                released["done"] = True
                self.release()

        present_ocr_dialog(
            pil_image,
            parent=None,
            on_dialog_shown=lambda dialog: dialog.connect("closed", lambda *_: release_once()),
            on_cancelled=release_once,
            auto_copy=True,
        )

    def on_shutdown(self, application):
        logging.info("Application shutdown started, cleaning temp directories…")
        for temp_dir in self.temp_dirs:
            try:
                if os.path.exists(temp_dir):
                    shutil.rmtree(temp_dir)
                    logging.debug(f"Deleted temp dir: {temp_dir}")
            except Exception as e:
                logging.warning(f"Failed to clean up temp dir {temp_dir}.", exception=e, show_exception=True)
        logging.info("Cleanup complete.")

def main(version: str) -> int:
    try:
        logging.info("Application starting…")
        loader = StdinImageLoader()
        image_path = loader.read_from_stdin()

        app = GradiaApp(version=version)
        app._stdin_image_path = image_path

        return app.run(sys.argv)

    except Exception as e:
        logging.critical("Application closed with an exception.", exception=e, show_exception=True)
        return 1



# Copyright (C) 2025 tfuxu, Alexander Vanhee
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Adw, Gio, Gtk, Gdk, GLib, GObject, Graphene

from gradia.constants import rootdir
from gradia.overlay.drawing_overlay import DrawingOverlay
from gradia.overlay.transparency_overlay import TransparencyBackground
from gradia.overlay.crop_overlay import CropOverlay
from gradia.overlay.drop_overlay import DropOverlay
from gradia.ui.widget.aspect_ratio_button import AspectRatioButton
from gradia.overlay.zoom_controller import ZoomController
from gradia.backend.ocr import OCR

@Gtk.Template(resource_path=f"{rootdir}/ui/image_stack.ui")
class ImageStack(Adw.Bin):
    __gtype_name__ = "GradiaImageStack"
    __gsignals__ = {
        "crop-toggled": (GObject.SignalFlags.RUN_FIRST, None, (bool,)),
        "zoom-changed": (GObject.SignalFlags.RUN_FIRST, None, (float,))
    }

    stack: Gtk.Stack = Gtk.Template.Child()

    main_overlay: Gtk.Overlay = Gtk.Template.Child()
    zoomable_widget: ZoomController = Gtk.Template.Child()
    picture_overlay: Gtk.Overlay = Gtk.Template.Child()
    drawing_overlay: DrawingOverlay = Gtk.Template.Child()

    picture: Gtk.Picture = Gtk.Template.Child()
    transparency_background: TransparencyBackground = Gtk.Template.Child()
    crop_overlay: CropOverlay = Gtk.Template.Child()

    erase_controls_revealer: Gtk.Revealer = Gtk.Template.Child()
    crop_controls_revealer: Gtk.Revealer = Gtk.Template.Child()
    right_controls_revealer: Gtk.Revealer = Gtk.Template.Child()

    drop_overlay: DropOverlay = Gtk.Template.Child()

    ocr_revealer = Gtk.Template.Child()

    crop_options_revealer: Gtk.Revealer = Gtk.Template.Child()
    confirm_crop_revealer: Gtk.Revealer = Gtk.Template.Child()
    back_button: Gtk.Revealer = Gtk.Template.Child()
    crop_button_revealer: Gtk.Revealer = Gtk.Template.Child()

    zoom_label: Gtk.Label = Gtk.Template.Child()
    zoom_out_button: Gtk.Button = Gtk.Template.Child()
    zoom_in_button: Gtk.Button = Gtk.Template.Child()
    reset_zoom_button: Gtk.Button = Gtk.Template.Child()
    zoom_info_revealer: Gtk.Revealer = Gtk.Template.Child()
    sidebar_revealer: Gtk.Revealer = Gtk.Template.Child()
    sidebar_button: Gtk.Button = Gtk.Template.Child()

    crop_enabled: bool = False
    crop_has_been_enabled: bool = False

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self._compact = False
        self._saved_crop_rect = None
        self._saved_aspect_ratio = None
        self._setup()

        self.connect("realize", self._on_realize)

    def _on_realize(self, widget):
        self.get_root().lookup_action("ocr").bind_property(
            "state",
            self.ocr_revealer,
            "reveal-child",
            GObject.BindingFlags.SYNC_CREATE,
            lambda binding, state: state.get_boolean(),
            None
        )

        OCR(self.get_root())

    def _get_ocr_action_state(self):
        return self.get_root().lookup_action("ocr").get_state()

    @GObject.Property(type=bool, default=False)
    def compact(self) -> bool:
        return self._compact

    @compact.setter
    def compact(self, value: bool) -> None:
        if self._compact != value:
            self._compact = value
            self._update_compact_ui()

    def _update_compact_ui(self) -> None:
        self.sidebar_button.set_visible(self._compact)
        zoom_level = self.zoomable_widget.get_property("zoom-level")
        self.zoom_label.set_visible(not self._compact)

    def set_erase_selected_visible(self, show: bool) -> None:
        self.erase_controls_revealer.set_reveal_child(show)

    def _setup(self) -> None:
        self.transparency_background.set_picture_reference(self.picture)
        self.crop_overlay.set_picture_reference(self.picture)
        self.crop_overlay.set_can_target(False)
        self.drawing_overlay.set_picture_reference(self.picture)
        self.drawing_overlay.set_erase_selected_revealer(self.erase_controls_revealer)
        self.right_controls_revealer.set_reveal_child(True)

        self.zoomable_widget.set_child_widgets(
            self.picture_overlay,
            self.drawing_overlay,
            self.crop_overlay,
            self.transparency_background,
            self.picture
        )

        self.zoomable_widget.connect("notify::zoom-level", self._on_zoom_level_changed)

        drop_target = Gtk.DropTarget.new(Gio.File, Gdk.DragAction.COPY)
        drop_target.set_preload(True)
        drop_target.connect("drop", self._on_file_dropped)
        self.drop_overlay.drop_target = drop_target

    def is_compact(self) -> bool:
        return self._compact

    def _on_zoom_level_changed(self, widget, pspec) -> None:
        zoom_level = widget.get_property("zoom-level")
        percentage = int(zoom_level * 100)
        self.zoom_label.set_visible(not self._compact)
        self.zoom_label.set_text(f"{percentage}%")
        self.emit("zoom-changed", zoom_level)

        self.zoom_out_button.set_sensitive(zoom_level > widget.min_zoom)
        self.zoom_in_button.set_sensitive(zoom_level < widget.max_zoom)

        self.zoom_info_revealer.set_reveal_child(zoom_level != 1)

    def _on_file_dropped(self, _target: Gtk.DropTarget, value: Gio.File, _x: int, _y: int) -> bool:
        uri = value.get_uri()
        if uri:
            window = self.get_root()
            action = window.lookup_action("load-drop") if window else None
            if action:
                action.activate(GLib.Variant('s', uri))
                return True
        return False

    def reset_crop_selection(self, silent=False) -> None:
        self.crop_overlay.set_crop_rectangle(0.0, 0.0, 1, 1)
        self.crop_overlay.aspect_ratio = 0
        self.crop_has_been_enabled = False
        if not silent:
            self.on_toggle_crop()

    def on_toggle_crop(self) -> None:
        self.crop_enabled = not self.crop_enabled
        self.crop_overlay.interactive = self.crop_enabled
        self.crop_overlay.set_can_target(self.crop_enabled)
        self.right_controls_revealer.set_reveal_child(not self.crop_enabled)
        self.right_controls_revealer.set_sensitive(not self.crop_enabled)
        self.confirm_crop_revealer.set_reveal_child(self.crop_enabled)
        self.back_button.set_reveal_child(self.crop_enabled)
        self.crop_button_revealer.set_reveal_child(not self.crop_enabled)
        self.ocr_revealer.set_reveal_child((not self.crop_enabled) and self._get_ocr_action_state())
        self.sidebar_revealer.set_reveal_child(not self.crop_enabled)
        self.zoomable_widget.disable_zoom = self.crop_enabled
        self.sidebar_button.set_sensitive(not self.crop_enabled)
        if not self._compact:
            self._show_sidebar(not self.crop_enabled)
        self.crop_options_revealer.set_reveal_child(self.crop_enabled)

        if self.crop_enabled:
            self._saved_crop_rect = self.crop_overlay.get_crop_rectangle()
            self._saved_aspect_ratio = self.crop_overlay.aspect_ratio

        if self.crop_enabled and not self.crop_has_been_enabled:
            self.crop_overlay.set_crop_rectangle(0.1, 0.1, 0.8, 0.8)
            self.crop_has_been_enabled = True

    def crop_back(self) -> None:
        if self.crop_enabled:
            if self._saved_crop_rect is not None:
                x, y, w, h = self._saved_crop_rect
                self.crop_overlay.set_crop_rectangle(x, y, w, h)
                self._saved_crop_rect = None
            else:
                self.crop_overlay.set_crop_rectangle(0.0, 0.0, 1, 1)

            if self._saved_aspect_ratio is not None:
                self.crop_overlay.aspect_ratio = self._saved_aspect_ratio
                self._saved_aspect_ratio = None
            else:
                self.crop_overlay.aspect_ratio = 0

            self.crop_has_been_enabled = False
            self.on_toggle_crop()

    def _show_sidebar(self, value):
        window = self.get_root()
        action = window.lookup_action("sidebar-shown")
        if action:
            action.activate(GLib.Variant('b', value))
            return True

    def set_aspect_ratio(self, ratio: float) -> None:
        if self.crop_enabled:
            self.crop_overlay.aspect_ratio = ratio

    def zoom_in(self, factor: float = 1.2) -> None:
        self.zoomable_widget.zoom_in(factor)

    def zoom_out(self, factor: float = 0.8) -> None:
        self.zoomable_widget.zoom_out(factor)

    def reset_zoom(self) -> None:
        self.zoomable_widget.reset_zoom()

    def set_zoom_level(self, zoom_level: float) -> None:
        self.zoomable_widget.zoom_level = zoom_level

    def get_zoom_level(self) -> float:
        return self.zoomable_widget.zoom_level

    def pan(self, dx, dy):
        self.zoomable_widget.pan(dx,dy)

    def on_image_loaded(self) -> None:
        self.reset_zoom()
        self._on_zoom_level_changed(self.zoomable_widget, None)
